#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import os
import tarfile
import uuid
import time
from collections import OrderedDict

from .... import options, tempobj
from ....compat import BytesIO, six
from ....utils import TEMP_TABLE_PREFIX
from ....errors import ODPSError


UDF_CLASS_NAME = 'PyOdpsFunc'

expr_to_sql = dict()
expr_deps = dict()
expr_ref_name = dict()


class ODPSContext(object):
    def __init__(self, odps, indent_size=2):
        self._odps = odps

        self._index = itertools.count(1)
        self._expr_alias = dict()
        self._compiled_exprs = dict()
        self._expr_raw_args = dict()

        self._col_index = itertools.count(1)
        self._need_alias_columns = dict()
        self._need_alias_column_indexes = dict()

        self._select_index = itertools.count(1)
        self._expr_ref_index = itertools.count(1)

        self._func_to_udfs = OrderedDict()
        self._registered_funcs = OrderedDict()
        self._func_to_functions = OrderedDict()
        self._func_to_resources = OrderedDict()

        self._indent_size = indent_size
        self._mapjoin_hints = []
        self._skewjoin_hints = []

        self._default_schema = odps.schema

        self._path_to_resources = dict()

        self._to_drops = []

    def next_select_id(self):
        return next(self._select_index)

    def new_alias_table_name(self):
        return 't%s' % next(self._index)

    def new_ref_name(self):
        return '@c%d' % next(self._expr_ref_index)

    def register_collection(self, expr):
        expr_id = id(expr)
        if expr_id in self._expr_alias:
            return self._expr_alias[expr_id]

        self._expr_alias[expr_id] = self.new_alias_table_name()
        return self._expr_alias[expr_id]

    def get_collection_alias(self, expr, create=False, silent=False):
        try:
            return self._expr_alias[id(expr)], False
        except KeyError as e:
            if create:
                return self.register_collection(expr), True
            if not silent:
                raise e

    def remove_collection_alias(self, expr):
        if id(expr) in self._expr_alias:
            del self._expr_alias[id(expr)]

    def add_expr_compiled(self, expr, compiled):
        self._compiled_exprs[id(expr)] = compiled

    def get_expr_compiled(self, expr):
        return self._compiled_exprs[id(expr)]

    def _gen_udf_name(self, func):
        identifier = getattr(func, "_identifier", None) or "udf"
        return 'pyodps_%s_%s_%s' % (
            identifier, int(time.time()), str(uuid.uuid4()).replace('-', '_')
        )

    def _gen_resource_name(self):
        return 'pyodps_res_%s_%s' % (int(time.time()), str(uuid.uuid4()).replace('-', '_'))

    def _gen_table_name(self):
        return '%s_%s_%s' % (TEMP_TABLE_PREFIX, int(time.time()),
                             str(uuid.uuid4()).replace('-', '_'))

    def register_udfs(self, func_to_udfs, func_to_resources):
        self._func_to_udfs = func_to_udfs
        for func in func_to_udfs.keys():
            self._registered_funcs[func] = self._gen_udf_name(func)
        self._func_to_resources = func_to_resources

    def get_udf(self, func):
        return self._registered_funcs[func]

    def get_udf_count(self):
        return len(self._registered_funcs)

    def prepare_resources(self, libraries):
        from ....models import Resource

        if libraries is None:
            return None

        ret_libs = []
        for lib in libraries:
            if isinstance(lib, Resource):
                ret_libs.append(lib)
                continue
            elif lib in self._path_to_resources:
                ret_libs.append(self._path_to_resources[lib])
                continue

            tarbinary = BytesIO()
            tar = tarfile.open(fileobj=tarbinary, mode='w:gz')
            if os.path.isfile(lib):
                with open(lib, 'rb') as fo:
                    finfo = tarfile.TarInfo('pyodps_files/' + os.path.basename(lib))
                    finfo.size = os.path.getsize(lib)
                    tar.addfile(finfo, fo)
            else:
                base_dir = os.path.dirname(os.path.abspath(lib))
                for root, dirs, files in os.walk(lib):
                    for f in files:
                        fpath = os.path.join(root, f)
                        rpath = os.path.relpath(fpath, base_dir).replace(os.path.sep, '/')
                        with open(fpath, 'rb') as fo:
                            finfo = tarfile.TarInfo(rpath)
                            finfo.size = os.path.getsize(fpath)
                            tar.addfile(finfo, fo)
            tar.close()

            res_name = self._gen_resource_name() + '.tar.gz'
            res = self._odps.create_resource(
                res_name, 'archive', fileobj=tarbinary.getvalue(), schema=self._default_schema, temp=True
            )
            if options.df.delete_udfs:
                tempobj.register_temp_resource(
                    self._odps, res_name, schema=self._default_schema
                )
                self._to_drops.append(res)
            self._path_to_resources[lib] = res
            ret_libs.append(res)
        return ret_libs

    def create_udfs(self, libraries=None):
        self._func_to_functions.clear()

        for func, udf in six.iteritems(self._func_to_udfs):
            udf_name = self._registered_funcs[func]
            py_resource = self._odps.create_resource(
                udf_name + '.py', 'py', file_obj=udf, schema=self._default_schema, temp=True
            )
            if options.df.delete_udfs:
                tempobj.register_temp_resource(
                    self._odps, udf_name + '.py', schema=self._default_schema
                )
                self._to_drops.append(py_resource)

            resources = [py_resource, ]
            if func in self._func_to_resources:
                for _, name, _, create, table_name in self._func_to_resources[func]:
                    if not create:
                        resources.append(name)
                    else:
                        res = self._odps.create_resource(
                            name, 'table', table_name=table_name, schema=self._default_schema, temp=True
                        )
                        if options.df.delete_udfs:
                            tempobj.register_temp_resource(self._odps, name, schema=self._default_schema)
                            self._to_drops.append(res)
                        resources.append(res)
            if libraries is not None:
                resources.extend(libraries)

            function_name = '{0}.{1}'.format(udf_name, UDF_CLASS_NAME)
            function = self._odps.create_function(
                udf_name, class_type=function_name, resources=resources, schema=self._default_schema
            )
            if options.df.delete_udfs:
                tempobj.register_temp_function(self._odps, udf_name, schema=self._default_schema)
                self._to_drops.append(function)

            self._func_to_functions[func] = function

    def add_need_alias_column(self, column):
        if id(column) in self._need_alias_column_indexes:
            return self._need_alias_column_indexes[id(column)]
        symbol = 'col_%s' % next(self._col_index)
        self._need_alias_columns[id(column)] = column
        self._need_alias_column_indexes[id(column)] = symbol
        return symbol

    def get_all_need_alias_column_symbols(self):
        for col_id, column in six.iteritems(self._need_alias_columns):
            symbol = self._need_alias_column_indexes[col_id]
            yield symbol, column


    #####################################
    # mem cache relatives
    #####################################

    def register_mem_cache_sql(self, expr, sql):
        expr_to_sql[expr._id] = sql

    def is_expr_mem_cached(self, expr):
        return expr._id in expr_to_sql

    def get_mem_cached_sql(self, expr):
        return expr_to_sql[expr._id]

    def register_mem_cache_dep(self, expr, dep):
        if expr._id not in expr_deps:
            expr_deps[expr._id] = set()
        expr_deps[expr._id].add(dep._id)

    def get_mem_cache_ref_name(self, expr, create=True):
        if not create:
            return expr_ref_name[expr._id]
        ref_name = expr_ref_name.get(expr._id, self.new_ref_name())
        if expr._id not in expr_ref_name:
            expr_ref_name[expr._id] = ref_name
        return ref_name

    def get_mem_cache_dep_sqls(self, *expr_ids):
        sqls = []
        fetched = dict()

        def h(eid):
            if eid in expr_deps:
                for dep_id in expr_deps[eid]:
                    if dep_id not in fetched:
                        h(dep_id)

            if eid not in fetched:
                origin_sql = expr_to_sql[eid]
                sql = '{0} := CACHE ON {1}'.format(
                    expr_ref_name[eid], origin_sql
                )
                sqls.append(sql)
                fetched[eid] = True

        for expr_id in expr_ids:
            h(expr_id)
        return sqls

    def _drop_silent(self, obj):
        try:
            obj.drop()
        except ODPSError:
            pass

    @property
    def default_schema(self):
        return self._default_schema

    @default_schema.setter
    def default_schema(self, value):
        self._default_schema = value

    def close(self):
        [self._drop_silent(it) for it in self._to_drops]



# encoding: utf-8
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from collections import namedtuple
from functools import partial

from ..expr.exporters import get_ml_input
from ..utils import parse_hist_repr
from ... import utils
from ...compat import StringIO, six
from ...df import DataFrame

try:
    import numpy as np
except ImportError:
    np = None

from ..enums import FieldRole
from odps.ml.expr.exporters import get_input_table_name, get_input_partitions, get_input_field_names


get_predicted_value_column = partial(get_input_field_names, field_role=FieldRole.PREDICTED_VALUE)


def get_y_table_name(expr, input_name):
    if not expr._params['yColName']:
        return None
    if not get_ml_input(expr, input_name):
        return get_input_table_name(expr, 'x')
    return get_input_table_name(expr, input_name)


def get_y_partitions(expr, input_name):
    if not expr._params['yColName']:
        return None
    if not get_ml_input(expr, input_name):
        return get_input_partitions(expr, 'x')
    return get_input_partitions(expr, input_name)


def get_histograms(expr, odps):
    hist_dict = dict()
    for rec in DataFrame(odps.get_table(expr.tables[0])).execute():
        col, bins = rec
        hist_dict[col] = parse_hist_repr(bins)
    return hist_dict


def get_pearson_result(expr, odps):
    return DataFrame(odps.get_table(expr.tables[0])).execute()[0][-1]


class TTestResult(object):
    def __init__(self, result):
        convs = dict(level='ConfidenceLevel', alpha='alpha', df='df', mean='mean', alternative='alternative', mu='mu',
                     p='p', t='t', stddev='stdDeviation', diff_mean='mean of the differences', x_mean='mean of x',
                     y_mean='mean of y', alter_hyp='AlternativeHypthesis')
        convs_comment = dict(x_mean='xMean', y_mean='yMean', x_dev='xDeviation', y_dev='yDeviation', cov='cov')

        self._items = set()

        for k, v in six.iteritems(convs):
            if v in result:
                setattr(self, k, result[v])
                self._items.add(k)

        if result.get('comment'):
            comm_values = json.loads('{' + result.get('comment') + '}')
            for k, v in six.iteritems(convs_comment):
                if v in comm_values:
                    setattr(self, k, comm_values[v])
                    self._items.add(k)

        self._items = sorted(self._items)
        convs.update(convs_comment)
        self._convs = convs

    def __repr__(self):
        buf = StringIO()
        space = 2 * max(len(it) for it in self._items)
        for name in self._items:
            buf.write('\n{0}{1}'.format(name.ljust(space), repr(getattr(self, name))))
        return 'TTestResult {{{0}\n}}'.format(utils.indent(buf.getvalue(), 2))

    def _repr_html_(self):
        buf = StringIO()
        buf.write('<table><tr><th>Field</th><th>Name</th><th>Value</th></tr>')
        for name in self._items:
            buf.write('<tr><td>{0}</td><td>{1}</td><td>{2}</td></tr>'.format(name, self._convs[name], getattr(self, name)))
        buf.write('</table>')
        return buf.getvalue()


def get_t_test_result(expr, odps):
    return TTestResult(json.loads(DataFrame(odps.get_table(expr.tables[0])).execute()[0][0]))


class ChiSquareTestResult(object):
    def __init__(self, result, correct_result, details):
        self.result = result
        self.correct_result = correct_result
        self.detail = details


def get_chisq_test_result(expr, odps):
    ChiSquareData = namedtuple('ChiSquareData', 'comment df p_value value')
    json_result = json.loads(DataFrame(odps.get_table(expr.tables.summary)).execute()[0][0].replace('p-value', 'p_value'))
    result = ChiSquareData(**json_result.get('Chi-Square'))
    correct_result = ChiSquareData(**json_result.get('Continity Correct Chi-Square'))

    fields = 'f0 f1 observed expected residuals'.split()
    details = []
    for rec in DataFrame(odps.get_table(expr.tables.detail)).execute():
        details.append(dict(zip(fields, rec.values)))

    return ChiSquareTestResult(result, correct_result, details)


def get_mat_result(expr, odps):
    MatResult = namedtuple('MatResult', 'matrix cols')
    cols = []
    mat = []
    for rec in DataFrame(odps.get_table(expr.tables[0])).execute():
        if not cols:
            cols = list(rec.keys()[1:])
        mat.append(rec.values[1:])
    if np:
        mat = np.matrix(mat)
    return MatResult(mat, cols)



# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libc.stdint cimport *
from libc.string cimport *


cdef extern from "../src/crc32.c":
    uint32_t crc32(uint32_t crc, const void *buf, size_t length) nogil
    uint32_t crc32c(uint32_t crc, const void *buf, size_t length) nogil


cdef class Checksum:

    def __cinit__(self, method='crc32c'):
        self._checksum = 0
        if method == 'crc32c':
            self.use_c = 1
        else:
            self.use_c = 0

    cdef void c_update_bool(self, bint val) nogil:
        cdef char retval
        retval = 1 if val else 0
        self.c_update(<char *>&retval, 1)

    cpdef update_bool(self, bint val):
        self.c_update_bool(val)

    cdef void c_update_int(self, int32_t val) nogil:
        self.c_update(<char *>&val, sizeof(int32_t))

    cpdef update_int(self, int32_t val):
        self.c_update(<char *>&val, sizeof(int32_t))

    cdef void c_update_long(self, int64_t val) nogil:
        self.c_update(<char *>&val, sizeof(int64_t))

    cpdef update_long(self, int64_t val):
        self.c_update(<char *>&val, sizeof(int64_t))

    cdef void c_update_float(self, float val) nogil:
        self.c_update(<char *>&val, sizeof(float))

    cpdef update_float(self, float val):
        self.c_update(<char *>&val, sizeof(float))

    cdef void c_update_double(self, double val) nogil:
        self.c_update(<char *>&val, sizeof(double))

    cpdef update_double(self, double val):
        self.c_update(<char *>&val, sizeof(double))

    cdef void c_update(self, const char *ptr, size_t length) nogil:
        if self.use_c:
            self._checksum = crc32c(self._checksum, ptr, length)
        else:
            self._checksum = crc32(self._checksum, ptr, length)

    cpdef update(self, bytes b):
        buf = bytearray(b)
        if self.use_c:
            self._checksum = crc32c(self._checksum, <const void*>(<char *>buf), len(buf))
        else:
            self._checksum = crc32(self._checksum, <const void*>(<char *>buf), len(buf))

    cdef uint32_t c_getvalue(self) nogil:
        return self._checksum

    cpdef uint32_t getvalue(self):
        return self._checksum

    cdef uint32_t c_setvalue(self, uint32_t val) nogil:
        self._checksum = val

    cpdef uint32_t setvalue(self, uint32_t val):
        self._checksum = val

    cdef void c_reset(self) nogil:
        self._checksum = 0

    cpdef reset(self):
        self._checksum = 0



# encoding: utf-8
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import collections
import re
import logging
import time
from json import JSONEncoder
from types import FunctionType, MethodType
from copy import deepcopy
from functools import partial
from threading import Thread
from collections import namedtuple

from .. import options
from ..compat import six, getargspec, Iterable
from ..types import Partition
from ..utils import camel_to_underline, TEMP_TABLE_PREFIX
from .enums import FieldContinuity, FieldRole

ML_PACKAGE_ROOT = 'odps.ml'
ML_INTERNAL_PACKAGE_ROOT = 'odps.internal.ml'
TEMP_MODEL_PREFIX = 'pm_'
TABLE_MODEL_PREFIX = 'otm_'
TEMP_TABLE_MODEL_PREFIX = TEMP_TABLE_PREFIX + 'otm_'
TABLE_MODEL_SEPARATOR = '__'

ML_ARG_PREFIX = '_mlattr_'

logger = logging.getLogger(__name__)


class KVConfig(object):
    def __init__(self, kv=':', item=','):
        self.kv_delimiter = kv
        self.item_delimiter = item

    def __repr__(self):
        return 'KVConfig(kv={0}, item={1})'.format(self.kv_delimiter, self.item_delimiter)

    def __eq__(self, other):
        return (self.kv_delimiter, self.item_delimiter) == (other.kv_delimiter, other.item_delimiter)


class MLFields(list):
    def __getslice__(self, i, j):
        return MLFields(super(MLFields, self).__getslice__(i, j))

    def _repr_html_(self):
        html = '<table>'
        html += '<thead><tr><th>Field</th><th>Type</th><th>Roles</th><th>Continuity</th></tr></thead>'
        html += '<tbody>'
        for f in sorted(self):
            html += '<tr>'
            html += '<td>%s</td>' % f.name
            html += '<td>%s</td>' % f._repr_type_()
            html += '<td>%s</td>' % f._repr_role_()
            html += '<td>%s</td>' % ('Continuous' if f.continuity == FieldContinuity.CONTINUOUS else 'Discrete')
            html += '</tr>'
        html += '</tbody></table>'
        return html


class MLField(object):
    """
    Represent table field definition

    :type continuity: FieldContinuity | None
    """
    _role_types = dict((str(e.name).upper(), e) for e in FieldRole)

    def __init__(self, name, field_type, role, continuity=None, is_append=False, is_partition=False, kv_config=None):
        self.name = name
        # field data type
        self.type = field_type.lower()
        # enum in FieldRole
        if role is None:
            role = set()
        self.role = set(role) if isinstance(role, Iterable) else set([role, ])
        # enum in FieldContinuity
        if continuity is not None:
            self.continuity = continuity
        elif self.type == 'double' or self.type == 'bigint':
            self.continuity = FieldContinuity.CONTINUOUS
        else:
            self.continuity = FieldContinuity.DISCRETE
        self.is_append = is_append
        self.is_partition = is_partition
        self.kv_config = kv_config

    def copy(self, role=None):
        if role is None:
            role = set()
        ret = deepcopy(self)
        new_roles = role if isinstance(role, Iterable) else set([role, ])
        ret.role = ret.role | new_roles
        return ret

    @classmethod
    def from_column(cls, column, role=None):
        return cls(column.name, str(column.type).lower(), role)

    @classmethod
    def from_string(cls, col_str):
        sparts = col_str.split(':')
        role = cls.translate_role_name(sparts[2].strip()) if len(sparts) > 2 else None
        return cls(sparts[0].strip(), sparts[1].strip().lower(), role)

    @classmethod
    def translate_role_name(cls, role_name):
        role_name = role_name.strip()
        if not role_name:
            return None
        if isinstance(role_name, FieldRole):
            return role_name
        try:
            role_name = str(role_name).strip().upper()
            return cls._role_types[role_name]
        except KeyError:
            raise KeyError('Role name ' + role_name + ' not registered. ' +
                           'Consider fixing the name or importing the right algorithm packages.')

    @classmethod
    def register_roles(cls, new_enums):
        cls._role_types.update(dict((str(e.name).upper(), e) for e in new_enums))

    def _repr_type_(self):
        return "KV('%s', '%s')" % (self.kv_config.kv_delimiter, self.kv_config.item_delimiter) if self.kv_config else self.type

    def _repr_role_(self):
        return '+'.join(v.name for v in self.role) if self.role else 'None'

    def __repr__(self):
        return '[%s]%s (%s) -> %s' % (self.continuity.name[0] if self.continuity is not None else 'N', self.name,
                                      self._repr_type_(), self._repr_role_())


def is_temp_table(table_name):
    return table_name.startswith(TEMP_TABLE_PREFIX)


def get_function_args(func):
    if isinstance(func, partial):
        arg_names = get_function_args(func.func)[len(func.args):]
        return [arg for arg in arg_names if arg not in func.keywords]
    elif not isinstance(func, (MethodType, FunctionType)) and hasattr(func, '__call__'):
        return get_function_args(getattr(func, '__call__'))
    else:
        return list(getargspec(func).args)


def import_class_member(path):
    import importlib

    package, func = path.rsplit('.', 1)

    outer_package = package.replace('$package_root', '.')
    try:
        imported = importlib.import_module(outer_package, __name__)
        return getattr(imported, func)
    except (ImportError, AttributeError):
        inter_package = package.replace('$package_root', '...internal.ml')
        if inter_package == outer_package:
            raise
        imported = importlib.import_module(inter_package, __name__)
        return getattr(imported, func)

HistogramOutput = namedtuple('HistogramOutput', 'hist bin_edges')


def parse_hist_repr(bins):
    hist, edges = [], []
    try:
        import numpy as np
    except ImportError:
        np = None
    for bar_str in bins.split(';'):
        bin_str, hist_str = bar_str.split(':')
        bin_str = bin_str.strip()
        if bin_str.endswith(']'):
            left, right = bin_str.strip('[]').split(',')
            edges.extend([float(left.strip()), float(right.strip())])
        else:
            left, _ = bin_str.strip('[)').split(',')
            edges.append(float(left.strip()))
        hist.append(float(hist_str.strip()))
    if np is not None:
        hist = np.array(hist)
        edges = np.array(edges)
    return HistogramOutput(hist=hist, bin_edges=edges)


def build_model_table_name(model_name, item_name):
    if model_name.startswith(TEMP_TABLE_PREFIX):
        model_name = model_name[len(TEMP_TABLE_PREFIX):]
        return ''.join((TEMP_TABLE_MODEL_PREFIX, model_name, TABLE_MODEL_SEPARATOR, item_name))
    else:
        return ''.join((TABLE_MODEL_PREFIX, model_name, TABLE_MODEL_SEPARATOR, item_name))



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
import threading
import time

from mars.deploy.kubernetes.web import K8SWebApplication

from .client import CUPID_APP_NAME
from .core import CupidServiceMixin

from .. import web as _web_plugin

del _web_plugin

logger = logging.getLogger(__name__)


class CupidWebApplication(CupidServiceMixin, K8SWebApplication):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        from cupid import context

        self._cupid_context = context()

    def _app_register_thread(self, internal_endpoint, app_endpoint):
        while True:
            if self._cupid_context is None:
                break
            kvstore = self._cupid_context.kv_store()
            kvstore[CUPID_APP_NAME] = json.dumps(dict(endpoint=internal_endpoint))

            self._cupid_context.register_application(CUPID_APP_NAME, app_endpoint)
            time.sleep(5)

    def start(self):
        import socket

        super().start()

        host_addr = socket.gethostbyname(socket.gethostname())
        internal_endpoint = "http://{0}:{1}".format(host_addr, self.mars_web.port)

        if os.environ.get("VM_ENGINE_TYPE") == "hyper":
            app_endpoint = socket.gethostname() + "-{}".format(self.mars_web.port)
        else:
            app_endpoint = internal_endpoint

        app_register_thread = threading.Thread(
            target=self._app_register_thread,
            args=(internal_endpoint, app_endpoint),
            daemon=True,
        )
        app_register_thread.start()

    def stop(self):
        super().stop()
        if self._cupid_context is not None:
            self._cupid_context.channel_client.stop()
            self._cupid_context = None


main = CupidWebApplication()

if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ...expr.expressions import Column, CollectionExpr


def change_input(expr, src_input, new_input, get_field, dag):
    traversed = set()
    list(expr.traverse(traversed=traversed, stop_cond=lambda x: x is src_input, unique=True))
    for p in filter(lambda x: x._node_id in traversed, dag.successors(src_input)):
        if isinstance(p, Column):
            col_name = p.source_name or p.name
            field = get_field(new_input, col_name)
            if p.is_renamed():
                field = field.rename(p.name)
            else:
                field = field.copy()
            if p is expr:
                p.substitute(src_input, new_input, dag=dag)
            else:
                parents = [it for it in dag.successors(p)
                           if it._node_id in traversed]
                dag.substitute(p, field, parents=parents)
        else:
            assert isinstance(p, CollectionExpr)
            p.substitute(src_input, new_input, dag=dag)


def copy_sequence(sequence, collection, dag=None):
    if dag is None:
        dag = sequence.to_dag(copy=False)

    traversed = set()
    copies = dict()

    def copy(node):
        if node._node_id not in copies:
            copies[node._node_id] = node.copy()
        copied = copies[node._node_id]
        if not dag.contains_node(copied):
            dag.add_node(copied)
        return copied

    for n in sequence.traverse(top_down=True, traversed=traversed,
                               stop_cond=lambda x: x is collection, unique=True):
        if n is sequence:
            copy(n)
            continue
        if n is collection:
            continue

        parents = [p for p in dag.successors(n) if p._node_id in traversed]
        for parent in parents:
            copy(parent).substitute(n, copy(n), dag=dag)

    return copies[sequence._node_id]


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .core import Container
from .project import Project
from .. import serializers, errors
from ..compat import six


class Projects(Container):
    __slots__ = "_odps_ref",

    marker = serializers.XMLNodeField('Marker')
    max_items = serializers.XMLNodeField('MaxItems')
    projects = serializers.XMLNodesReferencesField(Project, 'Project')

    def __init__(self, *args, **kwargs):
        self._odps_ref = None
        super(Projects, self).__init__(*args, **kwargs)

    def _get(self, item):
        return Project(client=self._client, _parent=self, name=item)

    def __contains__(self, item):
        if isinstance(item, six.string_types):
            project = self._get(item)
        elif isinstance(item, Project):
            project = item
        else:
            return False

        try:
            project.reload()
            return True
        except errors.NoSuchObject:
            return False

    def __iter__(self):
        return self.iterate()

    def iterate(
        self, owner=None, user=None, group=None, max_items=None, name=None,
        region_id=None, tenant_id=None, quota_nick_name=None, quota_type=None,
        quota_name=None,
    ):
        params = {
            'expectmarker': 'true',
            'name': name,
            'owner': owner,
            'user': user,
            'group': group,
            'maxitems': max_items,
            'region': region_id,
            'tenant': tenant_id,
            'quotanickname': quota_nick_name,
            'quota_name': quota_name,
            'quota_type': quota_type,
        }
        params = dict(
            (k, v) for k, v in params.items() if v is not None
        )

        def _it():
            last_marker = params.get('marker')
            if 'marker' in params and \
                    (last_marker is None or len(last_marker) == 0):
                return

            url = self.resource()
            resp = self._client.get(url, params=params)

            t = Projects.parse(self._client, resp, obj=self)
            params['marker'] = t.marker

            return t.projects

        while True:
            projects = _it()
            if projects is None:
                break
            for project in projects:
                yield project



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from .expressions import *
from .groupby import *
from . import utils
from .. import types
from ..utils import FunctionWrapper


class SequenceReduction(Scalar):
    __slots__ = '_unique',
    _args = '_input',

    @property
    def node_name(self):
        return self.__class__.__name__

    @property
    def source_name(self):
        if self._source_name:
            return self._source_name
        return self._input.name

    @property
    def name(self):
        if self._name:
            return self._name
        source_name = self.source_name
        if source_name:
            return '%s_%s' % (source_name, self.node_name.lower())

    def rename(self, new_name):
        if new_name == self._name:
            return self

        attr_values = dict((attr, getattr(self, attr, None))
                           for attr in utils.get_attrs(self))
        attr_values['_name'] = new_name
        new_reduction = type(self)(**attr_values)

        return new_reduction

    def output_type(self):
        return repr(self._value_type)

    def to_grouped_reduction(self, grouped):
        collection = next(n for n in self.input.traverse(top_down=True, unique=True)
                          if isinstance(n, CollectionExpr))
        if collection is not grouped._input:
            raise ExpressionError('Aggregation should be applied to %s, %s instead' % (
                grouped._input, collection))

        cls_name = 'Grouped%s' % self.__class__.__name__
        clz = globals()[cls_name]

        kwargs = dict((arg, getattr(self, arg, None)) for arg in utils.get_attrs(self)
                       if arg != '_cache_args')
        kwargs['_data_type'] = kwargs.pop('_value_type')
        if '_source_value_type' in kwargs:
            kwargs['_source_data_type'] = kwargs.pop('_source_value_type')
        kwargs['_grouped'] = grouped
        del kwargs['_value']

        return clz(**kwargs)

    @property
    def input(self):
        return self._input

    def accept(self, visitor):
        return visitor.visit_reduction(self)


class GroupedSequenceReduction(SequenceExpr):
    __slots__ = '_grouped', '_unique'
    _args = '_input', '_by'
    _extra_args = '_grouped',

    def _init(self, *args, **kwargs):
        self._init_attr('_grouped', None)
        self._init_attr('_by', None)
        super(GroupedSequenceReduction, self)._init(*args, **kwargs)
        assert self._grouped is not None
        if self._by is None:
            self._by = self._grouped._by

    @property
    def source_name(self):
        if self._source_name:
            return self._source_name
        return self._input.name

    @property
    def name(self):
        if self._name:
            return self._name
        source_name = self.source_name
        if source_name:
            return '%s_%s' % (source_name, self.node_name.lower())

    def _repr(self):
        expr = self._grouped.agg(self)[self.name]

        return expr._repr()

    @property
    def input(self):
        return self._input

    def accept(self, visitor):
        return visitor.visit_reduction(self)


class Min(SequenceReduction):
    __slots__ = ()


class GroupedMin(GroupedSequenceReduction):
    node_name = 'min'


class Max(SequenceReduction):
    __slots__ = ()


class GroupedMax(GroupedSequenceReduction):
    node_name = 'max'


class Count(SequenceReduction):
    __slots__ = ()


class GroupedCount(GroupedSequenceReduction):
    node_name = 'count'


class Sum(SequenceReduction):
    __slots__ = ()


class GroupedSum(GroupedSequenceReduction):
    node_name = 'sum'


class Var(SequenceReduction):
    __slots__ = '_ddof',


class GroupedVar(GroupedSequenceReduction):
    __slots__ = '_ddof',
    node_name = 'var'


class Std(SequenceReduction):
    __slots__ = '_ddof',


class GroupedStd(GroupedSequenceReduction):
    __slots__ = '_ddof',
    node_name = 'std'


class Mean(SequenceReduction):
    __slots__ = ()


class GroupedMean(GroupedSequenceReduction):
    node_name = 'mean'


class Quantile(SequenceReduction):
    __slots__ = '_prob',


class GroupedQuantile(GroupedSequenceReduction):
    __slots__ = '_prob',
    node_name = 'quantile'


class Moment(SequenceReduction):
    __slots__ = '_order', '_center'


class GroupedMoment(GroupedSequenceReduction):
    __slots__ = '_order', '_center'
    node_name = 'moment'


class Skewness(SequenceReduction):
    __slots__ = ()
    node_name = 'skew'


class GroupedSkewness(GroupedSequenceReduction):
    node_name = 'skew'


class Kurtosis(SequenceReduction):
    __slots__ = ()


class GroupedKurtosis(GroupedSequenceReduction):
    node_name = 'kurtosis'


class Median(SequenceReduction):
    __slots__ = ()


class GroupedMedian(GroupedSequenceReduction):
    node_name = 'median'


class Any(SequenceReduction):
    __slots__ = ()


class GroupedAny(GroupedSequenceReduction):
    node_name = 'any'


class All(SequenceReduction):
    __slots__ = ()


class GroupedAll(GroupedSequenceReduction):
    node_name = 'all'


class NUnique(SequenceReduction):
    _args = '_inputs',

    @property
    def source_name(self):
        if self._source_name:
            return self._source_name
        if len(self._inputs) == 1:
            return self._inputs[0].name

    @property
    def name(self):
        if self._name:
            return self._name
        source_name = self.source_name
        if source_name:
            return '%s_%s' % (source_name, self.node_name.lower())

    @property
    def input(self):
        return self._inputs[0]


class GroupedNUnique(GroupedSequenceReduction):
    node_name = 'nunique'
    _args = '_inputs', '_by'

    @property
    def source_name(self):
        if self._source_name:
            return self._source_name
        if len(self._inputs) == 1:
            return self._inputs[0].name

    @property
    def name(self):
        if self._name:
            return self._name
        source_name = self.source_name
        if source_name:
            return '%s_%s' % (source_name, self.node_name.lower())

    @property
    def input(self):
        return self._inputs[0]


class Cat(SequenceReduction):
    _args = '_input', '_by', '_sep', '_na_rep'

    def _init(self, *args, **kwargs):
        self._init_attr('_sep', None)
        self._init_attr('_na_rep', None)
        super(Cat, self)._init(*args, **kwargs)
        if self._sep is not None and not isinstance(self._sep, Scalar):
            self._sep = Scalar(_value=self._sep)
        if self._na_rep is not None and not isinstance(self._na_rep, Scalar):
            self._na_rep = Scalar(_value=self._na_rep)


class GroupedCat(GroupedSequenceReduction):
    _args = '_input', '_by', '_sep', '_na_rep'
    node_name = 'cat'

    def _init(self, *args, **kwargs):
        self._init_attr('_sep', None)
        self._init_attr('_na_rep', None)
        super(GroupedCat, self)._init(*args, **kwargs)
        if self._sep is not None and not isinstance(self._sep, Scalar):
            self._sep = Scalar(_value=self._sep)
        if self._na_rep is not None and not isinstance(self._na_rep, Scalar):
            self._na_rep = Scalar(_value=self._na_rep)


class ToList(SequenceReduction):
    __slots__ = ()


class GroupedToList(GroupedSequenceReduction):
    __slots__ = ()
    node_name = 'tolist'


class Aggregation(SequenceReduction):
    __slots__ = '_aggregator', '_func_args', '_func_kwargs', '_resources', '_raw_inputs'
    _args = '_inputs', '_collection_resources', '_by'
    node_name = 'Aggregation'

    def _init(self, *args, **kwargs):
        self._init_attr('_raw_inputs', None)
        super(Aggregation, self)._init(*args, **kwargs)

    @property
    def source_name(self):
        if self._source_name:
            return self._source_name
        if len(self._inputs) == 1:
            return self._inputs[0].name

    @property
    def raw_inputs(self):
        return self._raw_inputs or self._inputs

    @property
    def input(self):
        return self._inputs[0]

    @property
    def func(self):
        return self._aggregator

    @func.setter
    def func(self, f):
        self._aggregator = f

    @property
    def input_types(self):
        return [f.dtype for f in self._inputs]

    @property
    def raw_input_types(self):
        if self._raw_inputs:
            return [f.dtype for f in self._raw_inputs]
        return self.input_types

    def accept(self, visitor):
        visitor.visit_user_defined_aggregator(self)


class GroupedAggregation(GroupedSequenceReduction):
    __slots__ = '_aggregator', '_func_args', '_func_kwargs', '_resources', '_raw_inputs'
    _args = '_inputs', '_collection_resources', '_by'
    node_name = 'Aggregation'

    def _init(self, *args, **kwargs):
        self._init_attr('_raw_inputs', None)
        super(GroupedAggregation, self)._init(*args, **kwargs)

    @property
    def source_name(self):
        if self._source_name:
            return self._source_name
        if len(self._inputs) == 1:
            return self._inputs[0].name

    @property
    def raw_inputs(self):
        return self._raw_inputs or self._inputs

    @property
    def input(self):
        return self._inputs[0]

    @property
    def func(self):
        return self._aggregator

    @func.setter
    def func(self, f):
        self._aggregator = f

    @property
    def input_types(self):
        return [f.dtype for f in self._inputs]

    @property
    def raw_input_types(self):
        if self._raw_inputs:
            return [f.dtype for f in self._raw_inputs]
        return self.input_types

    def accept(self, visitor):
        visitor.visit_user_defined_aggregator(self)


def _extract_unique_input(expr):
    from .collections import DistinctCollectionExpr

    if isinstance(expr, Column):
        if isinstance(expr.input, DistinctCollectionExpr) and len(expr.input._unique_fields) == 1:
            return expr.input._unique_fields[0]
        else:
            return None


def _reduction(expr, output_cls, output_type=None, **kw):
    grouped_output_cls = globals()['Grouped%s' % output_cls.__name__]
    method_name = output_cls.__name__.lower()

    if isinstance(expr, CollectionExpr):
        columns = []

        for name in expr.schema.names:
            column = expr[name]
            if hasattr(column, method_name):
                columns.append(getattr(column, method_name)(**kw))

        return expr[columns]
    elif isinstance(expr, GroupBy):
        aggs = []

        # TODO: dynamic may need rebuilt
        for name in expr._to_agg.names:
            agg = expr[name]
            if hasattr(agg, method_name):
                aggs.append(getattr(agg, method_name)(**kw))

        return expr.agg(aggs)

    if output_type is None:
        output_type = expr._data_type

    if isinstance(expr, SequenceExpr):
        unique_input = _extract_unique_input(expr)
        if unique_input is not None:
            kw['_unique'] = True
            return _reduction(unique_input, output_cls, output_type=output_type, **kw)
        else:
            return output_cls(_value_type=output_type, _input=expr, **kw)
    elif isinstance(expr, SequenceGroupBy):
        return grouped_output_cls(_data_type=output_type, _input=expr.to_column(),
                                  _grouped=expr.input, **kw)


def min_(expr):
    """
    Min value

    :param expr:
    :return:
    """

    return _reduction(expr, Min)


def max_(expr):
    """
    Max value

    :param expr:
    :return:
    """

    return _reduction(expr, Max)


def count(expr):
    """
    Value counts

    :param expr:
    :return:
    """

    if isinstance(expr, SequenceExpr):
        unique_input = _extract_unique_input(expr)
        if unique_input:
            return nunique(unique_input).rename('count')
        else:
            return Count(_value_type=types.int64, _input=expr)
    elif isinstance(expr, SequenceGroupBy):
        return GroupedCount(_data_type=types.int64, _input=expr.to_column(),
                            _grouped=expr.input)
    elif isinstance(expr, CollectionExpr):
        return Count(_value_type=types.int64, _input=expr).rename('count')
    elif isinstance(expr, GroupBy):
        return GroupedCount(_data_type=types.int64, _input=expr.input,
                            _grouped=expr).rename('count')


def _stats_type(expr):
    if isinstance(expr, (SequenceExpr, SequenceGroupBy)):
        if expr._data_type == types.decimal:
            output_type = types.decimal
        else:
            output_type = types.float64
        return output_type


def var(expr, **kw):
    """
    Variance

    :param expr:
    :param ddof: degree of freedom
    :param kw:
    :return:
    """

    ddof = kw.get('ddof', kw.get('_ddof', 1))

    output_type = _stats_type(expr)
    return _reduction(expr, Var, output_type, _ddof=ddof)


def sum_(expr):
    """
    Sum value

    :param expr:
    :return:
    """

    output_type = None
    if isinstance(expr, (SequenceExpr, SequenceGroupBy)):
        if expr._data_type == types.boolean:
            output_type = types.int64
        else:
            output_type = expr._data_type
    return _reduction(expr, Sum, output_type)


def std(expr, **kw):
    """
    Standard deviation.

    :param expr:
    :param kw:
    :return:
    """

    ddof = kw.get('ddof', kw.get('_ddof', 1))

    output_type = _stats_type(expr)
    return _reduction(expr, Std, output_type, _ddof=ddof)


def mean(expr):
    """
    Arithmetic mean.

    :param expr:
    :return:
    """

    output_type = _stats_type(expr)
    return _reduction(expr, Mean, output_type)


def median(expr):
    """
    Median value.

    :param expr:
    :return:
    """

    output_type = _stats_type(expr)
    return _reduction(expr, Median, output_type)


def quantile(expr, prob=None, **kw):
    """
    Percentile value.

    :param expr:
    :param prob: probability or list of probabilities, in [0, 1]
    :return:
    """

    prob = kw.get('_prob', prob)
    output_type = _stats_type(expr)
    if isinstance(prob, (list, set)) and not isinstance(expr, GroupBy):
        output_type = types.List(output_type)
    return _reduction(expr, Quantile, output_type, _prob=prob)


def any_(expr):
    """
    Any is True.

    :param expr:
    :return:
    """

    output_type = types.boolean
    return _reduction(expr, Any, output_type)


def all_(expr):
    """
    All is True.

    :param expr:
    :return:
    """

    output_type = types.boolean
    return _reduction(expr, All, output_type)


def nunique(expr):
    """
    The distinct count.

    :param expr:
    :return:
    """
    output_type = types.int64

    if isinstance(expr, SequenceExpr):
        return NUnique(_value_type=output_type, _inputs=[expr])
    elif isinstance(expr, SequenceGroupBy):
        return GroupedNUnique(_data_type=output_type, _inputs=[expr.to_column()], _grouped=expr.input)
    elif isinstance(expr, CollectionExpr):
        unique_input = _extract_unique_input(expr)
        if unique_input:
            return nunique(unique_input)
        else:
            return NUnique(_value_type=types.int64, _inputs=expr._project_fields)
    elif isinstance(expr, GroupBy):
        if expr._to_agg:
            inputs = expr.input[expr._to_agg.names]._project_fields
        else:
            inputs = expr.input._project_fields
        return GroupedNUnique(_data_type=types.int64, _inputs=inputs,
                              _grouped=expr)


def _cat(expr, sep=None, na_rep=None):
    output_type = types.string
    return _reduction(expr, Cat, output_type, _sep=sep, _na_rep=na_rep)


def cat(expr, others=None, sep=None, na_rep=None):
    """
    Concatenate strings in sequence with given separator

    :param expr:
    :param others: other sequences
    :param sep: string or None, default None
    :param na_rep: string or None default None, if None, NA in the sequence are ignored
    :return:
    """

    if others is not None:
        from .strings import _cat as cat_str

        return cat_str(expr, others, sep=sep, na_rep=na_rep)

    return _cat(expr, sep=sep, na_rep=na_rep)


def moment(expr, order, central=False):
    """
    Calculate the n-th order moment of the sequence

    :param expr:
    :param order: moment order, must be an integer
    :param central: if central moments are to be computed.
    :return:
    """
    if not isinstance(order, six.integer_types):
        raise ValueError('Only integer-ordered moments are supported.')
    if order < 0:
        raise ValueError('Only non-negative orders are supported.')
    output_type = _stats_type(expr)
    return _reduction(expr, Moment, output_type, _order=order, _center=central)


def skew(expr):
    """
    Calculate skewness of the sequence

    :param expr:
    :return:
    """
    output_type = _stats_type(expr)
    return _reduction(expr, Skewness, output_type)


def kurtosis(expr):
    """
    Calculate kurtosis of the sequence

    :param expr:
    :return:
    """
    output_type = _stats_type(expr)
    return _reduction(expr, Kurtosis, output_type)


def aggregate(exprs, aggregator, rtype=None, resources=None, unique=False, args=(), **kwargs):
    name = None
    if isinstance(aggregator, FunctionWrapper):
        if aggregator.output_names:
            if len(aggregator.output_names) > 1:
                raise ValueError('Aggregate column has more than one name')
            name = aggregator.output_names[0]
        if aggregator.output_types:
            rtype = rtype or aggregator.output_types[0]
        aggregator = aggregator._func

    if rtype is None:
        rtype = utils.get_annotation_rtype(aggregator.getvalue)

    if not isinstance(exprs, Iterable):
        exprs = [exprs, ]
        if rtype is None:
            rtype = exprs[0].dtype

    if rtype is None:
        raise ValueError('rtype should be specified')
    output_type = types.validate_data_type(rtype)

    collection = None
    if len(exprs) > 0:
        for expr in exprs:
            coll = next(it for it in expr.traverse(top_down=True, unique=True)
                        if isinstance(it, CollectionExpr))
            if collection is None:
                collection = coll
            elif collection is not coll:
                raise ValueError('The sequences to aggregate should come from the same collection')

    collection_resources = utils.get_collection_resources(resources)
    if all(isinstance(expr, SequenceGroupBy) for expr in exprs):
        inputs = [expr.to_column() for expr in exprs]
        return GroupedAggregation(_inputs=inputs, _aggregator=aggregator,
                                  _data_type=output_type, _name=name,
                                  _func_args=args, _func_kwargs=kwargs, _resources=resources,
                                  _collection_resources=collection_resources,
                                  _grouped=exprs[0].input)
    else:
        if not unique and len(exprs) == 1:
            unique_input = _extract_unique_input(exprs[0])
            if unique_input:
                unique = True
                exprs = [unique_input]
        return Aggregation(_inputs=exprs, _aggregator=aggregator,
                           _value_type=output_type, _name=name,
                           _func_args=args, _func_kwargs=kwargs, _resources=resources,
                           _collection_resources=collection_resources, _unique=unique)


def agg(*args, **kwargs):
    return aggregate(*args, **kwargs)


def tolist(expr, **kwargs):
    """
    Pack all data in the sequence into a list
    :param expr:
    :param unique: make every elements in the sequence to be unique
    :return:
    """
    unique = kwargs.get('unique', kwargs.get('_unique', False))
    output_type = None

    if isinstance(expr, (SequenceExpr, SequenceGroupBy)):
        output_type = types.List(expr._data_type)
    return _reduction(expr, ToList, output_type, _unique=unique)


_number_sequence_methods = dict(
    var=var,
    std=std,
    mean=mean,
    moment=moment,
    skew=skew,
    kurtosis=kurtosis,
    kurt=kurtosis,
    median=median,
    sum=sum_,
    quantile=quantile,
)

_sequence_methods = dict(
    min=min_,
    max=max_,
    count=count,
    size=count,
    nunique=nunique,
    tolist=tolist,
)

number_sequences = [globals().get(repr(t).capitalize() + SequenceExpr.__name__)
                    for t in types.number_types()]

for number_sequence in number_sequences:
    utils.add_method(number_sequence, _number_sequence_methods)

utils.add_method(SequenceExpr, _sequence_methods)

StringSequenceExpr.sum = sum_
StringSequenceExpr.cat = cat
BooleanSequenceExpr.sum = sum_
BooleanSequenceExpr.any = any_
BooleanSequenceExpr.all = all_
SequenceExpr.aggregate = aggregate
SequenceExpr.agg = aggregate

number_sequences_groupby = [globals().get(repr(t).capitalize() + SequenceGroupBy.__name__)
                            for t in types.number_types()]

for number_sequence_groupby in number_sequences_groupby:
    utils.add_method(number_sequence_groupby, _number_sequence_methods)

utils.add_method(SequenceGroupBy, _sequence_methods)

StringSequenceGroupBy.sum = sum_
StringSequenceGroupBy.cat = _cat
BooleanSequenceGroupBy.sum = sum_
BooleanSequenceGroupBy.any = any_
BooleanSequenceGroupBy.all = all_
SequenceGroupBy.aggregate = aggregate
SequenceGroupBy.agg = aggregate

# add method to collection expression
utils.add_method(CollectionExpr, _number_sequence_methods)
utils.add_method(CollectionExpr, _sequence_methods)
CollectionExpr.size = count
CollectionExpr.any = any_
CollectionExpr.all = all_

# add method to GroupBy
utils.add_method(GroupBy, _number_sequence_methods)
utils.add_method(GroupBy, _sequence_methods)
GroupBy.size = count
GroupBy.any = any_
GroupBy.all = all_



# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

cimport cython
import decimal
import json
import warnings
from collections import OrderedDict
from cpython.datetime cimport import_datetime
from libc.stdint cimport *
from libc.string cimport *

from ...src.types_c cimport BaseRecord
from ...src.utils_c cimport CMillisecondsConverter
from ..pb.decoder_c cimport CDecoder
from ..checksum_c cimport Checksum

from ... import utils, types, compat, options
from ...errors import DatetimeOverflowError
from ...models import Record
from ...readers import AbstractRecordReader
from ...types import PartitionSpec
from ..wireconstants import ProtoWireConstants

cdef int64_t MAX_READ_SIZE_LIMIT = (1 << 63) - 1


cdef:
    uint32_t WIRE_TUNNEL_META_COUNT = ProtoWireConstants.TUNNEL_META_COUNT
    uint32_t WIRE_TUNNEL_META_CHECKSUM = ProtoWireConstants.TUNNEL_META_CHECKSUM
    uint32_t WIRE_TUNNEL_END_RECORD = ProtoWireConstants.TUNNEL_END_RECORD

cdef:
    int64_t BOOL_TYPE_ID = types.boolean._type_id
    int64_t DATETIME_TYPE_ID = types.datetime._type_id
    int64_t DATE_TYPE_ID = types.date._type_id
    int64_t STRING_TYPE_ID = types.string._type_id
    int64_t FLOAT_TYPE_ID = types.float_._type_id
    int64_t DOUBLE_TYPE_ID = types.double._type_id
    int64_t BIGINT_TYPE_ID = types.bigint._type_id
    int64_t BINARY_TYPE_ID = types.binary._type_id
    int64_t TIMESTAMP_TYPE_ID = types.timestamp._type_id
    int64_t INTERVAL_DAY_TIME_TYPE_ID = types.interval_day_time._type_id
    int64_t INTERVAL_YEAR_MONTH_TYPE_ID = types.interval_year_month._type_id
    int64_t DECIMAL_TYPE_ID = types.Decimal._type_id
    int64_t JSON_TYPE_ID = types.Json._type_id
    int64_t TIMESTAMP_NTZ_TYPE_ID = types.timestamp_ntz._type_id

cdef:
    object pd_timestamp = None
    object pd_timedelta = None

import_datetime()

cdef class BaseTunnelRecordReader:
    def __init__(self, object schema, object input_stream, columns=None, partition_spec=None):
        self._schema = schema
        if columns is None:
            self._columns = self._schema.columns
        else:
            self._columns = [self._schema[c] for c in columns]
        self._reader_schema = types.OdpsSchema(columns=self._columns)
        self._schema_snapshot = self._reader_schema.build_snapshot()
        self._n_columns = len(self._columns)
        self._overflow_date_as_none = options.tunnel.overflow_date_as_none
        self._struct_as_dict = options.struct_as_dict
        self._partition_vals = []

        partition_spec = PartitionSpec(partition_spec) if partition_spec is not None else None

        self._column_setters.resize(self._n_columns)
        for i in range(self._n_columns):
            data_type = self._schema_snapshot._col_types[i]
            if data_type == types.boolean:
                self._column_setters[i] = self._set_bool
            elif data_type == types.datetime:
                self._column_setters[i] = self._set_datetime
            elif data_type == types.date:
                self._column_setters[i] = self._set_date
            elif data_type == types.string:
                self._column_setters[i] = self._set_string
            elif data_type == types.float_:
                self._column_setters[i] = self._set_float
            elif data_type == types.double:
                self._column_setters[i] = self._set_double
            elif data_type in types.integer_types:
                self._column_setters[i] = self._set_bigint
            elif data_type == types.binary:
                self._column_setters[i] = self._set_string
            elif data_type == types.timestamp:
                self._column_setters[i] = self._set_timestamp
            elif data_type == types.timestamp_ntz:
                self._column_setters[i] = self._set_timestamp_ntz
            elif data_type == types.interval_day_time:
                self._column_setters[i] = self._set_interval_day_time
            elif data_type == types.interval_year_month:
                self._column_setters[i] = self._set_interval_year_month
            elif data_type == types.json:
                self._column_setters[i] = self._set_json
            elif isinstance(data_type, types.Decimal):
                self._column_setters[i] = self._set_decimal
            elif isinstance(data_type, (types.Char, types.Varchar)):
                self._column_setters[i] = self._set_string
            else:
                self._column_setters[i] = NULL
            if partition_spec is not None and self._columns[i].name in partition_spec:
                self._partition_vals.append((i, partition_spec[self._columns[i].name]))

        self._reader = CDecoder(input_stream)
        self._crc = Checksum()
        self._crccrc = Checksum()
        self._curr_cursor = 0
        self._read_limit = -1 if options.table_read_limit is None else options.table_read_limit
        self._mills_converter = CMillisecondsConverter()
        self._mills_converter_utc = CMillisecondsConverter(local_tz=False)
        self._to_date = utils.to_date

    def _mode(self):
        return 'c'

    @property
    def count(self):
        return self._curr_cursor

    cdef object _read_struct(self, object value_type):
        cdef:
            list res_list = [None] * len(value_type.field_types)
            int idx
        for idx, field_type in enumerate(value_type.field_types.values()):
            if not self._reader.read_bool():
                res_list[idx] = self._read_element(field_type._type_id, field_type)
        if self._struct_as_dict:
            return OrderedDict(zip(value_type.field_types.keys(), res_list))
        else:
            return value_type.namedtuple_type(*res_list)

    cdef object _read_element(self, int data_type_id, object data_type):
        if data_type_id == FLOAT_TYPE_ID:
            val = self._read_float()
        elif data_type_id == BIGINT_TYPE_ID:
            val = self._read_bigint()
        elif data_type_id == DOUBLE_TYPE_ID:
            val = self._read_double()
        elif data_type_id == STRING_TYPE_ID:
            val = self._read_string()
        elif data_type_id == BOOL_TYPE_ID:
            val = self._read_bool()
        elif data_type_id == DATETIME_TYPE_ID:
            val = self._read_datetime()
        elif data_type_id == BINARY_TYPE_ID:
            val = self._read_string()
        elif data_type_id == TIMESTAMP_TYPE_ID:
            val = self._read_timestamp()
        elif data_type_id == TIMESTAMP_NTZ_TYPE_ID:
            val = self._read_timestamp_ntz()
        elif data_type_id == INTERVAL_DAY_TIME_TYPE_ID:
            val = self._read_interval_day_time()
        elif data_type_id == INTERVAL_YEAR_MONTH_TYPE_ID:
            val = compat.Monthdelta(self._read_bigint())
        elif data_type_id == JSON_TYPE_ID:
            val = json.loads(self._read_string())
        elif data_type_id == DECIMAL_TYPE_ID:
            val = self._read_string()
        elif isinstance(data_type, (types.Char, types.Varchar)):
            val = self._read_string()
        elif isinstance(data_type, types.Array):
            val = self._read_array(data_type.value_type)
        elif isinstance(data_type, types.Map):
            keys = self._read_array(data_type.key_type)
            values = self._read_array(data_type.value_type)
            val = OrderedDict(zip(keys, values))
        elif isinstance(data_type, types.Struct):
            val = self._read_struct(data_type)
        else:
            raise IOError('Unsupported type %s' % data_type)
        return val

    cdef list _read_array(self, object value_type):
        cdef:
            uint32_t size
            int value_type_id = value_type._type_id
            object val

        size = self._reader.read_uint32()

        cdef list res = [None] * size
        for idx in range(size):
            if not self._reader.read_bool():
                res[idx] = self._read_element(value_type_id, value_type)
        return res

    cdef bytes _read_string(self):
        cdef bytes val = self._reader.read_string()
        self._crc.c_update(val, len(val))
        return val

    cdef float _read_float(self) except? -1.0 nogil:
        cdef float val = self._reader.read_float()
        self._crc.c_update_float(val)
        return val

    cdef double _read_double(self) except? -1.0 nogil:
        cdef double val = self._reader.read_double()
        self._crc.c_update_double(val)
        return val

    cdef bint _read_bool(self) except? False nogil:
        cdef bint val = self._reader.read_bool()
        self._crc.c_update_bool(val)
        return val

    cdef int64_t _read_bigint(self) except? -1 nogil:
        cdef int64_t val = self._reader.read_sint64()
        self._crc.c_update_long(val)
        return val

    cdef object _read_datetime(self):
        cdef int64_t val = self._reader.read_sint64()
        self._crc.c_update_long(val)
        try:
            return self._mills_converter.from_milliseconds(val)
        except DatetimeOverflowError:
            if not self._overflow_date_as_none:
                raise
            return None

    cdef object _read_date(self):
        cdef int64_t val = self._reader.read_sint64()
        self._crc.c_update_long(val)
        return self._to_date(val)

    cdef object _read_timestamp_base(self, bint ntz):
        cdef:
            int64_t val
            int32_t nano_secs
            CMillisecondsConverter converter

        global pd_timestamp, pd_timedelta

        if pd_timestamp is None:
            import pandas as pd
            pd_timestamp = pd.Timestamp
            pd_timedelta = pd.Timedelta

        val = self._reader.read_sint64()
        self._crc.c_update_long(val)
        nano_secs = self._reader.read_sint32()
        self._crc.c_update_int(nano_secs)

        if ntz:
            converter = self._mills_converter_utc
        else:
            converter = self._mills_converter

        try:
            return (
                pd_timestamp(converter.from_milliseconds(val * 1000))
                + pd_timedelta(nanoseconds=nano_secs)
            )
        except DatetimeOverflowError:
            if not self._overflow_date_as_none:
                raise
            return None

    cdef object _read_timestamp(self):
        return self._read_timestamp_base(False)

    cdef object _read_timestamp_ntz(self):
        return self._read_timestamp_base(True)

    cdef object _read_interval_day_time(self):
        cdef:
            int64_t val
            int32_t nano_secs
        global pd_timestamp, pd_timedelta

        if pd_timedelta is None:
            import pandas as pd
            pd_timestamp = pd.Timestamp
            pd_timedelta = pd.Timedelta
        val = self._reader.read_sint64()
        self._crc.c_update_long(val)
        nano_secs = self._reader.read_sint32()
        self._crc.c_update_int(nano_secs)
        return pd_timedelta(seconds=val, nanoseconds=nano_secs)

    cdef int _set_record_list_value(self, list record, int i, object value) except? -1:
        record[i] = self._schema_snapshot.validate_value(i, value, MAX_READ_SIZE_LIMIT)
        return 0

    cdef int _set_string(self, list record, int i) except? -1:
        cdef object val = self._read_string()
        self._set_record_list_value(record, i, val)
        return 0

    cdef int _set_float(self, list record, int i) except? -1:
        record[i] = self._read_float()
        return 0

    cdef int _set_double(self, list record, int i) except? -1:
        record[i] = self._read_double()
        return 0

    cdef int _set_bool(self, list record, int i) except? -1:
        record[i] = self._read_bool()
        return 0

    cdef int _set_bigint(self, list record, int i) except? -1:
        record[i] = self._read_bigint()
        return 0

    cdef int _set_datetime(self, list record, int i) except? -1:
        record[i] = self._read_datetime()
        return 0

    cdef int _set_date(self, list record, int i) except? -1:
        record[i] = self._read_date()
        return 0

    cdef int _set_decimal(self, list record, int i) except? -1:
        cdef bytes val = self._reader.read_string()
        self._crc.c_update(val, len(val))
        self._set_record_list_value(record, i, val)
        return 0

    cdef int _set_timestamp(self, list record, int i) except? -1:
        record[i] = self._read_timestamp()
        return 0

    cdef int _set_timestamp_ntz(self, list record, int i) except? -1:
        record[i] = self._read_timestamp_ntz()
        return 0

    cdef int _set_interval_day_time(self, list record, int i) except? -1:
        record[i] = self._read_interval_day_time()
        return 0

    cdef int _set_interval_year_month(self, list record, int i) except? -1:
        cdef int64_t val = self._read_bigint()
        self._set_record_list_value(record, i, compat.Monthdelta(val))
        return 0

    cdef int _set_json(self, list record, int i) except? -1:
        cdef bytes val = self._read_string()
        self._set_record_list_value(record, i, json.loads(val))
        return 0

    cpdef read(self):
        cdef:
            int index
            int checksum
            int idx_of_checksum
            int i
            int data_type_id
            object data_type
            BaseRecord record
            list rec_list

        if self._curr_cursor >= self._read_limit > 0:
            warnings.warn(
                'Number of lines read via tunnel already reaches the limitation.',
                RuntimeWarning,
            )
            return None

        record = Record(schema=self._reader_schema, max_field_size=MAX_READ_SIZE_LIMIT)
        rec_list = record._c_values

        while True:
            index = self._reader.read_field_number()

            if index == 0:
                continue
            if index == WIRE_TUNNEL_END_RECORD:
                checksum = <int32_t>self._crc.c_getvalue()
                if self._reader.read_uint32() != <uint32_t>checksum:
                    raise IOError('Checksum invalid')
                self._crc.c_reset()
                self._crccrc.c_update_int(checksum)
                break

            if index == WIRE_TUNNEL_META_COUNT:
                if self._curr_cursor != self._reader.read_sint64():
                    raise IOError('count does not match')
                idx_of_checksum = self._reader.read_field_number()
                if WIRE_TUNNEL_META_CHECKSUM != idx_of_checksum:
                    raise IOError('Invalid stream data.')
                if self._crccrc.c_getvalue() != self._reader.read_uint32():
                    raise IOError('Checksum invalid.')
                # if not self._reader.at_end():
                #     raise IOError('Expect at the end of stream, but not.')

                return

            if index > self._n_columns:
                raise IOError('Invalid protobuf tag. Perhaps the datastream '
                              'from server is crushed.')

            self._crc.c_update_int(index)

            i = index - 1
            if self._column_setters[i] != NULL:
                self._column_setters[i](self, rec_list, i)
            else:
                data_type = self._schema_snapshot._col_types[i]
                if isinstance(data_type, types.Array):
                    val = self._read_array(data_type.value_type)
                    rec_list[i] = self._schema_snapshot.validate_value(i, val, MAX_READ_SIZE_LIMIT)
                elif isinstance(data_type, types.Map):
                    keys = self._read_array(data_type.key_type)
                    values = self._read_array(data_type.value_type)
                    val = OrderedDict(zip(keys, values))
                    rec_list[i] = self._schema_snapshot.validate_value(i, val, MAX_READ_SIZE_LIMIT)
                elif isinstance(data_type, types.Struct):
                    val = self._read_struct(data_type)
                    rec_list[i] = self._schema_snapshot.validate_value(i, val, MAX_READ_SIZE_LIMIT)
                else:
                    raise IOError('Unsupported type %s' % data_type)

        for idx, val in self._partition_vals:
            rec_list[idx] = val

        self._curr_cursor += 1
        return record

    def reads(self):
        return self.__iter__()

    @property
    def n_bytes(self):
        return self._reader.position()

    def get_total_bytes(self):
        return self.n_bytes

    def close(self):
        if hasattr(self._schema, 'close'):
            self._schema.close()

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.close()


class TunnelRecordReader(BaseTunnelRecordReader, AbstractRecordReader):
    def __next__(self):
        record = self.read()
        if record is None:
            raise StopIteration
        return record

    next = __next__

    @property
    def schema(self):
        return self._schema


cdef int DECIMAL_FRAC_CNT = 2
cdef int DECIMAL_INTG_CNT = 4
cdef int DECIMAL_PREC_CNT = DECIMAL_INTG_CNT + DECIMAL_FRAC_CNT
cdef int DECIMAL_DIG_NUMS = 9
cdef int DECIMAL_FRAC_DIGS = DECIMAL_DIG_NUMS * DECIMAL_FRAC_CNT
cdef int DECIMAL_INTG_DIGS = DECIMAL_DIG_NUMS * DECIMAL_INTG_CNT
cdef int DECIMAL_PREC_DIGS = DECIMAL_DIG_NUMS * DECIMAL_PREC_CNT


@cython.cdivision(True)
cdef inline int32_t decimal_print_dig(
    char* buf, const int32_t* val, int count, bint tail = False
) nogil:
    cdef char* src = buf - count * DECIMAL_DIG_NUMS if tail else buf + 1
    cdef char* ret = src
    cdef char* ptr
    cdef int32_t i, data, r
    for i in range(count):
        ptr = buf
        data = val[i]
        while data != 0:
            r = data // 10
            ptr[0] = data - r * 10 + ord('0')
            if ptr[0] != ord('0') and (not tail or ret[0] == ord('0')):
                ret = ptr
            data = r
            ptr -= 1
        buf -= DECIMAL_DIG_NUMS
    return src - ret if src >= ret else ret - src


cpdef convert_legacy_decimal_bytes(bytes value, int32_t frac = 0):
    """
    Legacy decimal memory layout:
        int8_t  mNull;
        int8_t  mSign;
        int8_t  mIntg;
        int8_t  mFrac; only 0, 1, 2
        int32_t mData[6];
        int8_t mPadding[4]; //For Memory Align
    """
    if value is None:
        return None

    cdef const char *src_ptr = <const char *>value
    cdef bint is_null = src_ptr[0]
    cdef bint sign = src_ptr[1]
    cdef int mintg = src_ptr[2]
    cdef int mfrac = src_ptr[3]
    cdef const char *data = src_ptr + 4
    cdef int32_t dec_cnt

    cdef char buf[9 * (2 + 4) + 4]
    cdef char *buf_ptr = buf
    memset(buf_ptr, ord('0'), sizeof(buf))

    if is_null:  # pragma: no cover
        return None
    if mintg + mfrac == 0:  # IsZero
        buf[1] = ord(".")
        dec_cnt = 20  # "0.000000000000000000"
        if frac > 0:
            dec_cnt = dec_cnt if frac + 2 > dec_cnt else frac + 2
        else:
            dec_cnt = 1
        return decimal.Decimal(buf[0:dec_cnt].decode())

    cdef int32_t icnt = decimal_print_dig(
        buf_ptr + DECIMAL_INTG_DIGS, <const int32_t *>data + DECIMAL_FRAC_CNT, mintg
    )
    cdef char *start = buf_ptr + DECIMAL_INTG_DIGS + 1 - icnt if icnt > 0 else buf_ptr + DECIMAL_INTG_DIGS

    if sign:
        start -= 1
        start[0] = ord('-')

    cdef int32_t fcnt = decimal_print_dig(
        buf_ptr + DECIMAL_PREC_DIGS + 1, <const int32_t *>data, DECIMAL_FRAC_CNT, True
    )
    if frac <= DECIMAL_FRAC_DIGS:
        frac = frac if frac > 0 else 0
    else:
        frac = DECIMAL_FRAC_DIGS
    fcnt = max(fcnt, frac)
    buf[DECIMAL_INTG_DIGS + 1] = ord('.')

    dec_cnt = buf_ptr + DECIMAL_INTG_DIGS + 1 - start + (fcnt + 1 if fcnt > 0 else 0)
    return decimal.Decimal(start[0:dec_cnt].decode())



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
import pickle
import socket
import struct
import sys
import time
from concurrent.futures import ThreadPoolExecutor

from mars.lib.tblib import pickling_support
from ...compat import six
from ..utils import filter_partitions, check_partition_exist

pickling_support.install()

logger = logging.getLogger(__name__)

REQUEST_TYPE_READ_TABLE_DATA = 0
REQUEST_TYPE_WRITE_TABLE_DATA = 1
REQUEST_TYPE_ENUM_TABLE_PARTITIONS = 2
REQUEST_TYPE_CREATE_TABLE_DOWNLOAD_SESSION = 3
REQUEST_TYPE_CREATE_TABLE_UPLOAD_SESSION = 4
REQUEST_TYPE_COMMIT_TABLE_UPLOAD_SESSION = 5
REQUEST_TYPE_GET_KV = 6
REQUEST_TYPE_PUT_KV = 7
REQUEST_TYPE_TERMINATE_INSTANCE = 8
REQUEST_TYPE_GET_BEARER_TOKEN = 9
REQUEST_TYPE_REPORT_CONTAINER_STATUS = 10

CHUNK_BYTES_LIMIT = 64 * 1024**2
TRANSFER_BLOCK_SIZE = 64 * 1024**2
MAX_CHUNK_NUM = 512 * 1024**2


def _create_arrow_writer(sink, schema, **kwargs):
    import pyarrow as pa

    try:
        return pa.ipc.new_stream(sink, schema, **kwargs)
    except AttributeError:
        return pa.ipc.RecordBatchStreamWriter(sink, schema, **kwargs)


def _write_request_result(sock, success=True, result=None, exc_info=None):
    try:
        result_dict = {
            "status": success,
            "result": result,
            "exc_info": exc_info,
        }
        pickled = pickle.dumps({k: v for k, v in result_dict.items()})
        sock_out_file = sock.makefile("wb")
        sock_out_file.write(struct.pack("<I", len(pickled)))
        sock_out_file.write(pickled)
    finally:
        sock_out_file.flush()
        sock_out_file.close()


def _handle_read_table_data(sock):
    from cupid.io.table import TableSplit
    from cupid.errors import SubprocessStreamEOFError

    sock_out_file = sock.makefile("wb")
    ipc_writer = None
    try:
        (cmd_len,) = struct.unpack("<I", sock.recv(4))
        read_config = pickle.loads(sock.recv(cmd_len))
        min_rows = read_config.pop("min_rows", None)

        tsp = TableSplit(**read_config)
        logger.debug("Read split table, split index: %s", read_config["_split_index"])
        read_config = None

        if min_rows is None:
            reader = tsp.open_arrow_file_reader()
            while True:
                chunk = reader.read(TRANSFER_BLOCK_SIZE)
                if len(chunk) == 0:
                    break
                try:
                    sock_out_file.write(chunk)
                except (BrokenPipeError, SubprocessStreamEOFError):
                    break
                finally:
                    chunk = None
        else:
            reader = tsp.open_arrow_reader()
            nrows = 0
            while min_rows is None or nrows < min_rows:
                try:
                    batch = reader.read_next_batch()
                    nrows += batch.num_rows
                    if ipc_writer is None:
                        ipc_writer = _create_arrow_writer(sock_out_file, batch.schema)
                    ipc_writer.write_batch(batch)
                except StopIteration:
                    break
                finally:
                    batch = None
            if ipc_writer is not None:  # pragma: no branch
                ipc_writer.close()
                ipc_writer = None
    except:
        logger.exception("Failed to read table")
    finally:
        if ipc_writer is not None:
            ipc_writer.close()
        sock_out_file.flush()
        sock_out_file.close()


def _handle_write_table_data(sock):
    import pyarrow as pa
    from cupid.io.table.core import BlockWriter

    try:
        sock_in_file = sock.makefile("rb")

        (cmd_len,) = struct.unpack("<I", sock.recv(4))
        writer_config = pickle.loads(sock.recv(cmd_len))

        block_writer = BlockWriter(**writer_config)
        logger.debug(
            "Start writing table block, block id: %s", writer_config["_block_id"]
        )
        ipc_reader = pa.ipc.open_stream(sock_in_file)
        with block_writer.open_arrow_writer() as cupid_writer:
            arrow_writer = pa.RecordBatchStreamWriter(cupid_writer, ipc_reader.schema)
            while True:
                try:
                    batch = ipc_reader.read_next_batch()
                    arrow_writer.write_batch(batch)
                except StopIteration:
                    break
            arrow_writer.close()
        logger.debug(
            "Write table block finished, block id: %s", writer_config["_block_id"]
        )
        block_writer.commit()

        _write_request_result(sock, result={"block_id": writer_config["_block_id"]})
    except:
        logger.exception("Failed to read table")
        _write_request_result(sock, False, exc_info=sys.exc_info())


def _handle_enum_table_partitions(sock):
    try:
        (cmd_len,) = struct.unpack("<I", sock.recv(4))
        # dict with odps_params, table_name, partition
        task_config = pickle.loads(sock.recv(cmd_len))

        from odps import ODPS
        from odps.accounts import BearerTokenAccount
        from cupid import context

        cupid_ctx = context()

        odps_params = task_config["odps_params"]
        bearer_token = cupid_ctx.get_bearer_token()
        account = BearerTokenAccount(bearer_token)
        project = os.environ.get("ODPS_PROJECT_NAME", None) or odps_params["project"]
        endpoint = os.environ.get("ODPS_RUNTIME_ENDPOINT") or odps_params["endpoint"]
        o = ODPS(None, None, account=account, project=project, endpoint=endpoint)

        table = o.get_table(task_config["table_name"])
        partition_desc = task_config.get("partition")
        if not table.table_schema.partitions:
            _write_request_result(sock, result=None)
        elif partition_desc:
            if check_partition_exist(table, partition_desc):
                _write_request_result(sock, result=[partition_desc])
            else:
                parts = filter_partitions(o, list(table.partitions), partition_desc)
                _write_request_result(
                    sock, result=[str(pt.partition_spec) for pt in parts]
                )
        else:
            _write_request_result(
                sock, result=[str(pt.partition_spec) for pt in table.partitions]
            )
    except:
        logger.exception("Failed to create download session")
        _write_request_result(sock, False, exc_info=sys.exc_info())


def _handle_create_table_download_session(sock):
    try:
        (cmd_len,) = struct.unpack("<I", sock.recv(4))
        # dict with odps_params, table_name, partition, columns, worker_count, split_size, max_chunk_num
        session_config = pickle.loads(sock.recv(cmd_len))

        from odps import ODPS
        from odps.errors import ODPSError
        from odps.accounts import BearerTokenAccount
        from cupid import CupidSession, context
        from cupid.errors import CupidError
        from cupid.runtime import RuntimeContext

        if not RuntimeContext.is_context_ready():
            raise SystemError(
                "No Mars cluster found, please create via `o.create_mars_cluster`."
            )

        cupid_ctx = context()

        odps_params = session_config["odps_params"]
        bearer_token = cupid_ctx.get_bearer_token()
        account = BearerTokenAccount(bearer_token)
        project = os.environ.get("ODPS_PROJECT_NAME", None) or odps_params["project"]
        endpoint = os.environ.get("ODPS_RUNTIME_ENDPOINT") or odps_params["endpoint"]
        o = ODPS(None, None, account=account, project=project, endpoint=endpoint)
        cupid_session = CupidSession(o)

        split_size = session_config["split_size"]
        table_name = session_config["table_name"]
        data_src = o.get_table(table_name)
        if session_config.get("partition") is not None:
            data_src = data_src.get_partition(session_config["partition"])

        try:
            data_store_size = data_src.size
        except ODPSError:
            # fail to get data size, just ignore
            pass
        else:
            worker_count = session_config["worker_count"]
            if data_store_size < split_size and worker_count is not None:
                # data is too small, split as many as number of cores
                split_size = data_store_size // worker_count
                # at least 1M
                split_size = max(split_size, 1 * 1024**2)
                logger.debug(
                    "Input data size is too small, split_size is {}".format(split_size)
                )

        max_chunk_num = session_config["max_chunk_num"]
        columns = session_config["columns"]
        with_split_meta = session_config.get("with_split_meta_on_tile")

        logger.debug(
            "Start creating download session of table %s from cupid, columns %r",
            table_name,
            columns,
        )
        while True:
            try:
                download_session = cupid_session.create_download_session(
                    data_src,
                    split_size=split_size,
                    columns=columns,
                    with_split_meta=with_split_meta,
                )
                break
            except CupidError:
                logger.debug(
                    "The number of splits exceeds 100000, split_size is {}".format(
                        split_size
                    )
                )
                if split_size >= max_chunk_num:
                    raise
                else:
                    split_size *= 2

        ret_data = {
            "splits": download_session.splits,
            "split_size": split_size,
        }
        _write_request_result(sock, result=ret_data)
    except:
        logger.exception("Failed to create download session")
        _write_request_result(sock, False, exc_info=sys.exc_info())


def _handle_create_table_upload_session(sock):
    try:
        (cmd_len,) = struct.unpack("<I", sock.recv(4))
        # dict with odps_params, table_name
        session_config = pickle.loads(sock.recv(cmd_len))

        from odps import ODPS
        from odps.accounts import BearerTokenAccount
        from cupid import CupidSession, context
        from cupid.runtime import RuntimeContext

        if not RuntimeContext.is_context_ready():
            raise SystemError(
                "No Mars cluster found, please create via `o.create_mars_cluster`."
            )
        cupid_ctx = context()

        odps_params = session_config["odps_params"]
        bearer_token = cupid_ctx.get_bearer_token()
        account = BearerTokenAccount(bearer_token)
        project = os.environ.get("ODPS_PROJECT_NAME", None) or odps_params["project"]
        endpoint = os.environ.get("ODPS_RUNTIME_ENDPOINT") or odps_params["endpoint"]
        o = ODPS(None, None, account=account, project=project, endpoint=endpoint)
        cupid_session = CupidSession(o)

        data_src = o.get_table(session_config["table_name"])

        logger.debug("Start creating upload session from cupid.")
        upload_session = cupid_session.create_upload_session(data_src)

        ret_data = {
            "handle": upload_session.handle,
        }
        _write_request_result(sock, result=ret_data)
    except:
        logger.exception("Failed to create upload session")
        _write_request_result(sock, False, exc_info=sys.exc_info())


def _handle_commit_table_upload_session(sock):
    try:
        (cmd_len,) = struct.unpack("<I", sock.recv(4))
        # dict with odps_params, table_name, cupid_handle, blocks, overwrite
        commit_config = pickle.loads(sock.recv(cmd_len))

        from odps import ODPS
        from odps.accounts import BearerTokenAccount
        from cupid import CupidSession, context
        from cupid.runtime import RuntimeContext
        from cupid.io.table import CupidTableUploadSession

        if not RuntimeContext.is_context_ready():
            raise SystemError(
                "No Mars cluster found, please create via `o.create_mars_cluster`."
            )
        cupid_ctx = context()

        odps_params = commit_config["odps_params"]
        bearer_token = cupid_ctx.get_bearer_token()
        account = BearerTokenAccount(bearer_token)
        project = os.environ.get("ODPS_PROJECT_NAME", None) or odps_params["project"]
        endpoint = os.environ.get("ODPS_RUNTIME_ENDPOINT") or odps_params["endpoint"]
        o = ODPS(None, None, account=account, project=project, endpoint=endpoint)
        cupid_session = CupidSession(o)

        table = o.get_table(commit_config["table_name"])
        upload_session = CupidTableUploadSession(
            session=cupid_session,
            table_name=table.name,
            project_name=table.project.name,
            handle=commit_config["cupid_handle"],
            blocks=commit_config["blocks"],
        )
        upload_session.commit(overwrite=commit_config["overwrite"])

        _write_request_result(sock)
    except:
        logger.exception("Failed to commit upload session")
        _write_request_result(sock, False, exc_info=sys.exc_info())


def _handle_get_kv(sock):
    try:
        (cmd_len,) = struct.unpack("<I", sock.recv(4))
        # dict with key
        cmd_body = pickle.loads(sock.recv(cmd_len))

        from cupid.runtime import RuntimeContext

        if not RuntimeContext.is_context_ready():
            logger.warning("Cupid context not ready")
            value = None
        else:
            from cupid import context

            cupid_kv = context().kv_store()
            value = cupid_kv.get(cmd_body["key"])

        ret_data = {
            "value": value,
        }
        _write_request_result(sock, result=ret_data)
    except:
        logger.exception("Failed to get kv value")
        _write_request_result(sock, False, exc_info=sys.exc_info())


def _handle_put_kv(sock):
    try:
        (cmd_len,) = struct.unpack("<I", sock.recv(4))
        # dict with key
        cmd_body = pickle.loads(sock.recv(cmd_len))

        from cupid.runtime import RuntimeContext

        if not RuntimeContext.is_context_ready():
            logger.warning("Cupid context not ready")
        else:
            from cupid import context

            cupid_kv = context().kv_store()
            cupid_kv[cmd_body["key"]] = cmd_body["value"]

        _write_request_result(sock)
    except:
        logger.exception("Failed to put kv value")
        _write_request_result(sock, False, exc_info=sys.exc_info())


def _handle_terminate_instance(sock):
    from cupid.runtime import context, RuntimeContext
    from odps import ODPS
    from odps.accounts import BearerTokenAccount

    try:
        (cmd_len,) = struct.unpack("<I", sock.recv(4))
        # dict with key
        cmd_body = pickle.loads(sock.recv(cmd_len))

        instance_id = cmd_body["instance_id"]

        if not RuntimeContext.is_context_ready():
            logger.warning("Cupid context not ready")
        else:
            bearer_token = context().get_bearer_token()
            account = BearerTokenAccount(bearer_token)
            project = os.environ["ODPS_PROJECT_NAME"]
            endpoint = os.environ["ODPS_RUNTIME_ENDPOINT"]
            o = ODPS(None, None, account=account, project=project, endpoint=endpoint)

            o.stop_instance(instance_id)
    except:
        logger.exception("Failed to put kv value")
        _write_request_result(sock, False, exc_info=sys.exc_info())


def _handle_get_bearer_token(sock):
    try:
        (cmd_len,) = struct.unpack("<I", sock.recv(4))
        # dict with odps_params, table_name, cupid_handle, blocks, overwrite
        _commit_config = pickle.loads(sock.recv(cmd_len))  # noqa: F841

        from cupid import context

        bearer_token = context().get_bearer_token()
        _write_request_result(sock, result={"token": bearer_token})
    except:
        logger.exception("Failed to get bearer token")
        _write_request_result(sock, False, exc_info=sys.exc_info())


def _handle_report_container_status(sock):
    try:
        (cmd_len,) = struct.unpack("<I", sock.recv(4))
        # dict with odps_params, table_name, cupid_handle, blocks, overwrite
        cmd_body = pickle.loads(sock.recv(cmd_len))

        status = cmd_body["status"]
        message = cmd_body["message"]
        progress = cmd_body["progress"]
        timeout = cmd_body["timeout"]

        from cupid import context

        context().report_container_status(status, message, progress, timeout=timeout)
    except:
        logger.exception("Failed to get bearer token")
        _write_request_result(sock, False, exc_info=sys.exc_info())


_request_handlers = {
    REQUEST_TYPE_READ_TABLE_DATA: _handle_read_table_data,
    REQUEST_TYPE_WRITE_TABLE_DATA: _handle_write_table_data,
    REQUEST_TYPE_ENUM_TABLE_PARTITIONS: _handle_enum_table_partitions,
    REQUEST_TYPE_CREATE_TABLE_DOWNLOAD_SESSION: _handle_create_table_download_session,
    REQUEST_TYPE_CREATE_TABLE_UPLOAD_SESSION: _handle_create_table_upload_session,
    REQUEST_TYPE_COMMIT_TABLE_UPLOAD_SESSION: _handle_commit_table_upload_session,
    REQUEST_TYPE_GET_KV: _handle_get_kv,
    REQUEST_TYPE_PUT_KV: _handle_put_kv,
    REQUEST_TYPE_TERMINATE_INSTANCE: _handle_terminate_instance,
    REQUEST_TYPE_GET_BEARER_TOKEN: _handle_get_bearer_token,
    REQUEST_TYPE_REPORT_CONTAINER_STATUS: _handle_report_container_status,
}


def _handle_requests(sock):
    while True:
        try:
            (req_type,) = struct.unpack("<I", sock.recv(4))
            if req_type in _request_handlers:
                _request_handlers[req_type](sock)
            else:
                sock.close()
                break
        except ConnectionAbortedError:
            break


def _prepare_channel(channel_file):
    while not os.path.exists(channel_file):
        time.sleep(1)
    try:
        with open(channel_file, "r") as env_file:
            envs = json.loads(env_file.read())
    except:
        time.sleep(1)
        with open(channel_file, "r") as env_file:
            envs = json.loads(env_file.read())

    from cupid import context

    os.environ.update(envs)
    context()
    odps_envs = {
        "ODPS_BEARER_TOKEN": os.environ["BEARER_TOKEN_INITIAL_VALUE"],
        "ODPS_ENDPOINT": os.environ["ODPS_RUNTIME_ENDPOINT"],
    }
    os.environ.update(odps_envs)
    logger.info("Started channel for Cupid Server.")


def run_cupid_service(channel_file, sock_file=None, pool_size=None):
    _prepare_channel(channel_file)

    pool = ThreadPoolExecutor(pool_size)

    sock_file = sock_file or os.environ["CUPID_SERVICE_SOCKET"]
    logger.warning("Starting Cupid Service with socket %s", sock_file)
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock.bind(sock_file)
    sock.listen(1)

    while True:
        conn, _ = sock.accept()
        pool.submit(_handle_requests, conn)


class CupidServiceClient:
    def __init__(self, sock_file=None):
        self._sock_file = sock_file or os.environ["CUPID_SERVICE_SOCKET"]
        self._sock = None

    def __del__(self):
        self.close()

    def close(self):
        if self._sock is not None:
            self._sock.close()
            self._sock = None

    @property
    def sock(self):
        if self._sock is None or self._sock._closed:
            if not os.path.exists(self._sock_file):
                raise OSError("Socket file does not exist")
            self._sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self._sock.connect(self._sock_file)
        return self._sock

    def _send_cupid_service_request(self, req_type, arg_object):
        ser = pickle.dumps(arg_object)

        sock_out_file = self.sock.makefile("wb")
        sock_out_file.write(struct.pack("<I", req_type))
        sock_out_file.write(struct.pack("<I", len(ser)))
        sock_out_file.write(ser)
        sock_out_file.flush()

    def _fetch_cupid_service_result(self):
        (ret_len,) = struct.unpack("<I", self.sock.recv(4))
        ret = pickle.loads(self.sock.recv(ret_len))

        if not ret["status"]:
            six.reraise(*ret["exc_info"])
        else:
            return ret.get("result")

    def read_table_data(self, split_config, min_rows=None):
        import pyarrow as pa

        split_config = split_config.copy()
        split_config["min_rows"] = min_rows
        self._send_cupid_service_request(REQUEST_TYPE_READ_TABLE_DATA, split_config)

        sock_in_file = self.sock.makefile("rb")
        batches = []
        try:
            if min_rows is None:
                reader = pa.RecordBatchStreamReader(sock_in_file)
                return reader.read_all()
            else:
                ipc_reader = pa.ipc.open_stream(sock_in_file)
                while True:
                    try:
                        batches.append(ipc_reader.read_next_batch())
                    except StopIteration:
                        break
                return pa.Table.from_batches(batches)
        finally:
            batches[:] = []
            sock_in_file.close()

    def write_table_data(self, writer_config, to_store_data, write_batch_size=None):
        import pyarrow as pa
        from odps.mars_extension.utils import convert_pandas_object_to_string
        from odps.tunnel.io.types import odps_schema_to_arrow_schema

        writer_config = writer_config.copy()
        self._send_cupid_service_request(REQUEST_TYPE_WRITE_TABLE_DATA, writer_config)

        sock_out_file = self.sock.makefile("wb")

        batch_size = write_batch_size or 1024
        batch_idx = 0
        batch_data = to_store_data[
            batch_size * batch_idx : batch_size * (batch_idx + 1)
        ]
        batch_data = convert_pandas_object_to_string(batch_data)
        schema = odps_schema_to_arrow_schema((writer_config["_table_schema"]))
        arrow_writer = _create_arrow_writer(sock_out_file, schema)
        while len(batch_data) > 0:
            batch = pa.RecordBatch.from_pandas(
                batch_data, schema=schema, preserve_index=False
            )
            arrow_writer.write_batch(batch)
            batch_idx += 1
            batch_data = to_store_data[
                batch_size * batch_idx : batch_size * (batch_idx + 1)
            ]
        arrow_writer.close()
        sock_out_file.flush()

        return self._fetch_cupid_service_result()

    def enum_table_partitions(self, odps_params, table_name, partition=None):
        cmd_pack = {
            "odps_params": odps_params,
            "table_name": table_name,
            "partition": partition,
        }
        self._send_cupid_service_request(REQUEST_TYPE_ENUM_TABLE_PARTITIONS, cmd_pack)
        return self._fetch_cupid_service_result()

    def create_table_download_session(
        self,
        odps_params,
        table_name,
        partition,
        columns,
        worker_count=None,
        split_size=None,
        max_chunk_num=None,
        with_split_meta_on_tile=False,
    ):
        cmd_pack = {
            "odps_params": odps_params,
            "table_name": table_name,
            "partition": partition,
            "columns": columns,
            "worker_count": worker_count,
            "split_size": split_size or CHUNK_BYTES_LIMIT,
            "max_chunk_num": max_chunk_num or MAX_CHUNK_NUM,
            "with_split_meta_on_tile": with_split_meta_on_tile,
        }
        self._send_cupid_service_request(
            REQUEST_TYPE_CREATE_TABLE_DOWNLOAD_SESSION, cmd_pack
        )
        result = self._fetch_cupid_service_result()
        return result["splits"], result["split_size"]

    def create_table_upload_session(self, odps_params, table_name):
        cmd_pack = {
            "odps_params": odps_params,
            "table_name": table_name,
        }
        self._send_cupid_service_request(
            REQUEST_TYPE_CREATE_TABLE_UPLOAD_SESSION, cmd_pack
        )
        result = self._fetch_cupid_service_result()
        return result["handle"]

    def commit_table_upload_session(
        self, odps_params, table_name, cupid_handle, blocks, overwrite=True
    ):
        cmd_pack = {
            "odps_params": odps_params,
            "table_name": table_name,
            "cupid_handle": cupid_handle,
            "blocks": blocks,
            "overwrite": overwrite,
        }
        self._send_cupid_service_request(
            REQUEST_TYPE_COMMIT_TABLE_UPLOAD_SESSION, cmd_pack
        )
        self._fetch_cupid_service_result()

    def get_kv(self, key):
        cmd_pack = {"key": key}
        self._send_cupid_service_request(REQUEST_TYPE_GET_KV, cmd_pack)
        result = self._fetch_cupid_service_result()
        return result["value"]

    def put_kv(self, key, value):
        cmd_pack = {"key": key, "value": value}
        self._send_cupid_service_request(REQUEST_TYPE_PUT_KV, cmd_pack)
        self._fetch_cupid_service_result()

    def terminate_instance(self, instance_id):
        cmd_pack = {"instance_id": instance_id}
        self._send_cupid_service_request(REQUEST_TYPE_TERMINATE_INSTANCE, cmd_pack)
        self._fetch_cupid_service_result()

    def get_bearer_token(self):
        self._send_cupid_service_request(REQUEST_TYPE_GET_BEARER_TOKEN, {})
        result = self._fetch_cupid_service_result()
        return result["token"]

    def report_container_status(self, status, message, progress, timeout=-1):
        self._send_cupid_service_request(
            REQUEST_TYPE_REPORT_CONTAINER_STATUS,
            {
                "status": status,
                "message": message,
                "progress": progress,
                "timeout": timeout,
            },
        )
        self._fetch_cupid_service_result()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import weakref
import inspect

from ..compat import six


class ObjectCache(object):
    def __init__(self):
        self._caches = weakref.WeakValueDictionary()

    @staticmethod
    def _get_cache_class(cls):
        if hasattr(cls, '_cache_class'):
            cls._cache_class = ObjectCache._get_cache_class(cls._cache_class)
            return cls._cache_class
        return cls

    def _fetch(self, cache_key):
        client, parent, _, _ = cache_key
        if parent is None:
            return self._caches.get(cache_key)

        ancestor = getattr(parent, '_parent')
        parent_cls = self._get_cache_class(type(parent))
        if parent_cls is None:
            return None
        name = getattr(parent, 'name', parent_cls.__name__.lower())

        parent_cache_key = client, ancestor, parent_cls, name

        if self._fetch(parent_cache_key):
            return self._caches.get(cache_key)

    def _get_cache(self, cls, **kw):
        kwargs = dict(kw)
        parent = kwargs.pop('parent', None) or kwargs.pop('_parent', None)
        name = kwargs.pop(getattr(cls, '_cache_name_arg', 'name'), None)
        client = kwargs.pop('client', None) or kwargs.pop('_client', None)
        cache_cls = self._get_cache_class(cls)

        cache_key = client, parent, cache_cls, name
        obj = None
        if name is not None:
            obj = self._fetch(cache_key)
            if obj is not None:
                if not frozenset(kwargs).issubset(obj.__slots__):
                    obj = None
                else:
                    for k, v in six.iteritems(kwargs):
                        setattr(obj, k, v)
            if obj is not None:
                return cache_key, obj

        return cache_key, obj

    def cache_lazyload(self, func, cls, **kwargs):
        cache_key, obj = self._get_cache(cls, **kwargs)

        if obj is not None:
            return obj

        obj = func(cls, **kwargs)

        if not hasattr(cls, '_filter_cache'):
            self._caches[cache_key] = obj
        elif cls._filter_cache(func, **obj.extract(**kwargs)):
            self._caches[cache_key] = obj
        return obj

    def cache_container(self, func, cls, **kwargs):
        parent = kwargs.get('parent') or kwargs.get('_parent')
        client = kwargs.get('client') or kwargs.get('_client')
        name = cls.__name__.lower()

        cache_key = client, parent, cls, name
        if name is not None:
            obj = self._fetch(cache_key)
            if obj is not None:
                return obj

        obj = func(cls, **kwargs)
        self._caches[cache_key] = obj
        return obj

    def del_item_cache(self, obj, item):
        from .core import LazyLoad

        item = obj[item]

        client = getattr(item, '_client')
        parent = getattr(item, '_parent')
        name = item._name()

        if name is not None:
            clz = self._get_cache_class(type(item))
            cache_key = client, parent, clz, name
            if cache_key in self._caches:
                # make sure original object will be reloaded
                obj = self._caches[cache_key]
                if isinstance(obj, LazyLoad):
                    obj.reset()

                del self._caches[cache_key]


_object_cache = ObjectCache()


def cache(func):
    def inner(cls, **kwargs):
        bases = [base.__name__ for base in inspect.getmro(cls)]
        if 'LazyLoad' in bases:
            return _object_cache.cache_lazyload(func, cls, **kwargs)
        elif 'Container' in bases:
            return _object_cache.cache_container(func, cls, **kwargs)

        return func(cls, **kwargs)

    inner.__name__ = func.__name__
    inner.__doc__ = func.__doc__
    return inner


def del_cache(func):
    def inner(obj, item):
        if func.__name__ == '__delitem__':
            _object_cache.del_item_cache(obj, item)
        return func(obj, item)

    inner.__name__ = func.__name__
    inner.__doc__ = func.__doc__
    inner._cache_maker = True
    return inner


def cache_parent(cls):
    cls._cache_class = cls.__bases__[0]
    return cls



# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libc.stdint cimport *
from libc.string cimport *
from libcpp.vector cimport vector

import ctypes
import numpy as np
cimport numpy as np

from ..checksum_c cimport Checksum
from .block_decoder_c cimport Decoder
from .util_c cimport *
from . import errno
from ..wireconstants import ProtoWireConstants
from ... import types, options

try:
    import pandas as pd
except ImportError:
    pd = None

cdef:
    uint32_t WIRE_TUNNEL_META_COUNT = ProtoWireConstants.TUNNEL_META_COUNT
    uint32_t WIRE_TUNNEL_META_CHECKSUM = ProtoWireConstants.TUNNEL_META_CHECKSUM
    uint32_t WIRE_TUNNEL_END_RECORD = ProtoWireConstants.TUNNEL_END_RECORD

    int BD_SUCCESS = errno.BD_SUCCESS
    int BD_BUFFER_EXHAUSTED = errno.BD_BUFFER_EXHAUSTED
    int BD_CHECKSUM_INVALID = errno.BD_CHECKSUM_INVALID
    int BD_COUNT_NOT_MATCH = errno.BD_COUNT_NOT_MATCH
    int BD_INVALID_STREAM_DATA = errno.BD_INVALID_STREAM_DATA
    int BD_INVALID_PB_TAG = errno.BD_INVALID_PB_TAG


cdef class TunnelPandasReader:
    def __init__(self, object schema, object input_stream, columns=None):
        self._schema = schema
        if columns is None:
            self._columns = self._schema.columns
        else:
            self._columns = [self._schema[c] for c in columns]
        self._reader_schema = types.OdpsSchema(columns=self._columns)
        self._schema_snapshot = self._reader_schema.build_snapshot()
        self._n_columns = len(self._columns)

        self._mem_cache_size = options.tunnel.pd_mem_cache_size

        self._stream = input_stream

        self._row_mem_ptr = 0
        self._row_checksum = 0
        self._cur_cursor = 0
        self._mem_cache_bound = 0
        self._mem_cache = None
        self._crc = Checksum()
        self._crccrc = Checksum()

        self._scan_schema()

    cdef void _scan_schema(self):
        self._nogil_readers.resize(self._n_columns)
        self._use_no_gil = 1
        for i in range(self._n_columns):
            self._nogil_readers[i] = NULL
            data_type = self._schema_snapshot._col_types[i]
            if data_type == types.boolean:
                self._nogil_readers[i] = self._read_bool
            elif data_type == types.datetime:
                self._use_no_gil = 0
            elif data_type == types.string:
                self._use_no_gil = 0
            elif data_type == types.float_:
                self._nogil_readers[i] = self._read_float
            elif data_type == types.double:
                self._nogil_readers[i] = self._read_double
            elif data_type in types.integer_types:
                self._nogil_readers[i] = self._read_int64
            elif data_type == types.binary:
                self._use_no_gil = 0
            elif data_type == types.timestamp:
                self._use_no_gil = 0
            elif data_type == types.interval_day_time:
                self._use_no_gil = 0
            elif data_type == types.interval_year_month:
                self._use_no_gil = 0
            elif isinstance(data_type, types.Decimal):
                self._use_no_gil = 0
            elif isinstance(data_type, (types.Char, types.Varchar)):
                self._use_no_gil = 0
            else:
                self._use_no_gil = 0

    @property
    def row_mem_ptr(self):
        return self._row_mem_ptr

    @property
    def mem_cache_bound(self):
        return self._mem_cache_bound

    cpdef reset_positions(self, object cache, int cache_size):
        cdef char *cache_ptr
        cdef uint64_t cache_ptr_int

        self._mem_cache_bound = cache_size
        self._row_mem_ptr = 0
        self._row_ptr = 0
        self._crc.c_setvalue(self._row_checksum)
        self._mem_cache = cache

        if isinstance(cache, ctypes.Array):
            cache_ptr_int = ctypes.addressof(cache)
            cache_ptr = <char *>cache_ptr_int
        else:
            cache_ptr = cache
        self._decoder = Decoder()
        self._decoder.init(cache_ptr, self._mem_cache_bound)

    cpdef refill_cache(self):
        cdef bytearray new_mem_cache = bytearray(self._mem_cache_size)
        cdef object new_mem_view = memoryview(new_mem_cache)
        cdef object old_mem_view
        cdef int left_size = 0
        cdef int read_size = 0

        if self._mem_cache is not None:
            old_mem_view = memoryview(self._mem_cache)
            left_size = self._mem_cache_bound - self._row_mem_ptr
            if left_size:
                new_mem_view[0:left_size] = old_mem_view[self._row_mem_ptr:]

        read_size = self._stream.readinto(new_mem_view[left_size:])
        self.reset_positions(new_mem_cache, left_size + read_size)
        return read_size

    cdef void _read_bool(self, ArrayVariantPtrs &aptr, int row) nogil:
        cdef bint val
        val = self._decoder.read_bool()
        if self._decoder.get_last_error():
            return
        aptr.v.pbool_array[row] = val
        self._crc.c_update_bool(val)

    cdef void _read_int64(self, ArrayVariantPtrs &aptr, int row) nogil:
        cdef int64_t val
        val = self._decoder.read_sint64()
        if self._decoder.get_last_error():
            return
        aptr.v.pl_array[row] = val
        self._crc.c_update_long(val)

    cdef void _read_float(self, ArrayVariantPtrs &aptr, int row) nogil:
        cdef float val
        val = self._decoder.read_float()
        if self._decoder.get_last_error():
            return
        aptr.v.pflt_array[row] = val
        self._crc.c_update_float(val)

    cdef void _read_double(self, ArrayVariantPtrs &aptr, int row) nogil:
        cdef double val
        val = self._decoder.read_double()
        if self._decoder.get_last_error():
            return
        aptr.v.pdbl_array[row] = val
        self._crc.c_update_double(val)

    cdef int _fill_ndarrays_nogil(self, vector[ArrayVariantPtrs] &col_ptrs, int start_row, int limit) nogil:
        cdef int idx
        cdef int field_num
        cdef int idx_of_checksum
        cdef int rowid = start_row
        cdef int32_t checksum

        while rowid < limit:
            while True:
                field_num = self._decoder.read_field_number()
                if self._decoder.get_last_error() == BD_BUFFER_EXHAUSTED:
                    return rowid - start_row

                if field_num == 0:
                    continue
                if field_num == WIRE_TUNNEL_END_RECORD:
                    checksum = <int32_t>self._crc.c_getvalue()
                    if self._decoder.read_uint32() != <uint32_t>checksum:
                        if self._decoder.get_last_error() == BD_BUFFER_EXHAUSTED:
                            return rowid - start_row
                        self._decoder.set_last_error(BD_CHECKSUM_INVALID)
                        return rowid - start_row
                    self._crc.c_reset()
                    self._crccrc.c_update_int(checksum)
                    break
                if field_num == WIRE_TUNNEL_META_COUNT:
                    if self._cur_cursor != self._decoder.read_sint64():
                        if self._decoder.get_last_error() == BD_BUFFER_EXHAUSTED:
                            return rowid - start_row
                        self._decoder.set_last_error(BD_COUNT_NOT_MATCH)
                        return rowid - start_row
                    idx_of_checksum = self._decoder.read_field_number()
                    if self._decoder.get_last_error() == BD_BUFFER_EXHAUSTED:
                        return rowid - start_row
                    if WIRE_TUNNEL_META_CHECKSUM != idx_of_checksum:
                        self._decoder.set_last_error(BD_INVALID_STREAM_DATA)
                        return rowid - start_row
                    if self._crccrc.c_getvalue() != self._decoder.read_uint32():
                        if self._decoder.get_last_error() == BD_BUFFER_EXHAUSTED:
                            return rowid - start_row
                        self._decoder.set_last_error(BD_CHECKSUM_INVALID)
                        return rowid - start_row
                    self._row_mem_ptr = self._decoder.position()
                    return rowid - start_row

                if field_num > self._n_columns:
                    self._decoder.set_last_error(BD_INVALID_PB_TAG)
                    return rowid - start_row

                self._crc.c_update_int(field_num)

                idx = field_num - 1
                self._nogil_readers[idx](self, col_ptrs[idx], rowid)
                if self._decoder.get_last_error() == BD_BUFFER_EXHAUSTED:
                    return rowid - start_row
            self._row_mem_ptr = self._decoder.position()
            self._row_checksum = self._crc.c_getvalue()
            rowid += 1
            self._cur_cursor += 1

        return rowid - start_row

    cpdef int readinto(self, object buffers, object columns=None, int limit=-1) except? -1:
        """
        Read data into an existing buffer. The ``buffers`` variable can be a list or a dict
        of numpy arrays, or a Pandas DataFrame. The argument ``columns`` determines the subset
        and order of data.
        
        Currently only ``bigint``, ``float``, ``double`` and ``boolean`` are supported.
        
        :param buffers: data buffer to read, can be numpy array, dict or pandas DataFrame
        :param columns: column names to read
        :param limit: total number of records to read
        :return: number of records read
        """
        cdef:
            int i
            int filled
            int filled_total = 0
            int fetch_count = 0

            vector[ArrayVariantPtrs] col_ptrs
            int64_t[:] int_mmap
            float[:] flt_mmap
            double[:] dbl_mmap
            np.ndarray[np.npy_bool, ndim=1, cast=True] bool_array

            dict col_dict

        if not self._use_no_gil:
            raise NotImplementedError('Currently complex types are not supported.')

        col_dict = dict()
        if isinstance(buffers, dict):
            col_dict = buffers
        elif pd and isinstance(buffers, pd.DataFrame):
            for col_name in buffers:
                col_dict[col_name] = buffers[col_name].values
        elif columns is not None:
            for col_name, buf in zip(columns, buffers):
                col_dict[col_name] = buf
        else:
            for col, buf in zip(self._columns, buffers):
                col_dict[col.name] = buf

        if limit < 0:
            limit = 0x7fffffff
            for buf in col_dict.values():
                limit = min(limit, len(buf))

        col_ptrs.resize(self._n_columns)
        for i in range(self._n_columns):
            col_name = self._columns[i].name
            data_type = self._schema_snapshot._col_types[i]
            if data_type == types.float_:
                flt_mmap = col_dict[col_name]
                col_ptrs[i].v.pflt_array = &flt_mmap[0]
            elif data_type == types.double:
                dbl_mmap = col_dict[col_name]
                col_ptrs[i].v.pdbl_array = &dbl_mmap[0]
            elif data_type in types.integer_types:
                int_mmap = col_dict[col_name]
                col_ptrs[i].v.pl_array = &int_mmap[0]
            elif data_type == types.boolean:
                bool_array = col_dict[col_name]
                col_ptrs[i].v.pbool_array = <np.npy_bool *>bool_array.data

        while filled_total < limit:
            fetch_count = self.refill_cache()
            if fetch_count == 0:
                break
            self._decoder.set_last_error(BD_SUCCESS)
            filled = self._fill_ndarrays_nogil(col_ptrs, filled_total, limit)
            if self._decoder.get_last_error() != BD_SUCCESS:
                if self._decoder.get_last_error() == BD_CHECKSUM_INVALID:
                    raise IOError('Checksum invalid')
                elif self._decoder.get_last_error() == BD_INVALID_STREAM_DATA:
                    raise IOError('Invalid stream data')
                elif self._decoder.get_last_error() == BD_COUNT_NOT_MATCH:
                    raise IOError('Count not match')
                elif self._decoder.get_last_error() == BD_INVALID_PB_TAG:
                    raise IOError('Invalid protobuf tag. Perhaps the datastream '
                                  'from server is crushed.')
            filled_total += filled

        return filled_total

    def read(self, columns=None, limit=-1):
        """
        Read data into a pandas DataFrame. If pandas is not installed, a dict will be returned instead.
        The argument ``columns`` determines the subset and order of data.

        Currently only ``bigint``, ``float``, ``double`` and ``boolean`` are supported.

        :param columns: column names to read
        :param limit: total number of records to read
        :return: number of records read
        """
        if not columns:
            columns = self._schema.names

        buf = dict()
        results = dict()
        buf_len = options.tunnel.pd_row_cache_size

        for col in columns:
            results[col] = []
            col_type = self._schema[col].type
            if col_type == types.float_:
                buf[col] = np.empty((buf_len,), dtype=np.float32)
            elif col_type == types.double:
                buf[col] = np.empty((buf_len,), dtype=np.float64)
            elif col_type in types.integer_types:
                buf[col] = np.empty((buf_len,), dtype=np.int64)
            elif col_type == types.boolean:
                buf[col] = np.empty((buf_len,), dtype=np.bool_)

        read_all = limit < 0
        while read_all or limit > 0:
            read_cols = self.readinto(buf, columns, limit)
            if read_all:
                if read_cols == 0:
                    break
            else:
                limit -= read_cols
            for col in buf:
                results[col].append(np.copy(buf[col]))

        merged = dict()
        for col in results:
            if len(results[col]) > 1:
                merged[col] = np.concatenate(results[col])
            else:
                merged[col] = results[col][0]

        if pd:
            return pd.DataFrame(merged, columns=columns)
        else:
            return merged

    def close(self):
        if hasattr(self._schema, 'close'):
            self._schema.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import time

import cython
from cpython.datetime cimport import_datetime
from libc.stdint cimport *
from libc.string cimport *
from libcpp.vector cimport vector

from ...src.types_c cimport BaseRecord, SchemaSnapshot
from ..checksum_c cimport Checksum
from ..pb.encoder_c cimport CEncoder

from ... import types, compat, utils
from ...compat import six
from ...src.utils_c cimport CMillisecondsConverter
from ..pb.wire_format import WIRETYPE_VARINT as PY_WIRETYPE_VARINT, \
    WIRETYPE_FIXED32 as PY_WIRETYPE_FIXED32,\
    WIRETYPE_FIXED64 as PY_WIRETYPE_FIXED64,\
    WIRETYPE_LENGTH_DELIMITED as PY_WIRETYPE_LENGTH_DELIMITED
from ..wireconstants import ProtoWireConstants

cdef:
    uint32_t WIRETYPE_VARINT = PY_WIRETYPE_VARINT
    uint32_t WIRETYPE_FIXED32 = PY_WIRETYPE_FIXED32
    uint32_t WIRETYPE_FIXED64 = PY_WIRETYPE_FIXED64
    uint32_t WIRETYPE_LENGTH_DELIMITED = PY_WIRETYPE_LENGTH_DELIMITED

cdef:
    uint32_t WIRE_TUNNEL_META_COUNT = ProtoWireConstants.TUNNEL_META_COUNT
    uint32_t WIRE_TUNNEL_META_CHECKSUM = ProtoWireConstants.TUNNEL_META_CHECKSUM
    uint32_t WIRE_TUNNEL_END_RECORD = ProtoWireConstants.TUNNEL_END_RECORD

cdef:
    int64_t BOOL_TYPE_ID = types.boolean._type_id
    int64_t DATETIME_TYPE_ID = types.datetime._type_id
    int64_t DATE_TYPE_ID = types.date._type_id
    int64_t STRING_TYPE_ID = types.string._type_id
    int64_t FLOAT_TYPE_ID = types.float_._type_id
    int64_t DOUBLE_TYPE_ID = types.double._type_id
    int64_t BIGINT_TYPE_ID = types.bigint._type_id
    int64_t BINARY_TYPE_ID = types.binary._type_id
    int64_t TIMESTAMP_TYPE_ID = types.timestamp._type_id
    int64_t INTERVAL_DAY_TIME_TYPE_ID = types.interval_day_time._type_id
    int64_t INTERVAL_YEAR_MONTH_TYPE_ID = types.interval_year_month._type_id
    int64_t DECIMAL_TYPE_ID = types.Decimal._type_id
    int64_t JSON_TYPE_ID = types.Json._type_id
    int64_t TIMESTAMP_NTZ_TYPE_ID = types.timestamp_ntz._type_id

import_datetime()

cdef vector[int] data_type_to_wired_type

data_type_to_wired_type.resize(128, -1)
data_type_to_wired_type[BOOL_TYPE_ID] = WIRETYPE_VARINT
data_type_to_wired_type[DATETIME_TYPE_ID] = WIRETYPE_VARINT
data_type_to_wired_type[DATE_TYPE_ID] = WIRETYPE_VARINT
data_type_to_wired_type[BIGINT_TYPE_ID] = WIRETYPE_VARINT
data_type_to_wired_type[INTERVAL_YEAR_MONTH_TYPE_ID] = WIRETYPE_VARINT
data_type_to_wired_type[FLOAT_TYPE_ID] = WIRETYPE_FIXED32
data_type_to_wired_type[DOUBLE_TYPE_ID] = WIRETYPE_FIXED64
data_type_to_wired_type[STRING_TYPE_ID] = WIRETYPE_LENGTH_DELIMITED
data_type_to_wired_type[BINARY_TYPE_ID] = WIRETYPE_LENGTH_DELIMITED
data_type_to_wired_type[DECIMAL_TYPE_ID] = WIRETYPE_LENGTH_DELIMITED
data_type_to_wired_type[TIMESTAMP_TYPE_ID] = WIRETYPE_LENGTH_DELIMITED
data_type_to_wired_type[INTERVAL_DAY_TIME_TYPE_ID] = WIRETYPE_LENGTH_DELIMITED
data_type_to_wired_type[JSON_TYPE_ID] = WIRETYPE_LENGTH_DELIMITED
data_type_to_wired_type[TIMESTAMP_NTZ_TYPE_ID] = WIRETYPE_LENGTH_DELIMITED


cdef class ProtobufRecordWriter:

    def __init__(self, object output, buffer_size=None):
        self.DEFAULT_BUFFER_SIZE = 4096
        self._output = output
        self._n_total = 0
        self._buffer_size = buffer_size or self.DEFAULT_BUFFER_SIZE
        self._encoder = CEncoder(self._buffer_size)

    cpdef _re_init(self, output):
        self._encoder = CEncoder(self._buffer_size)
        self._output = output
        self._n_total = 0

    def _mode(self):
        return 'c'

    cpdef flush(self):
        if self._encoder.position() > 0:
            data = self._encoder.tostring()
            self._output.write(data)
            self._n_total += self._encoder.position()
            self._encoder = CEncoder(self._buffer_size)

    cpdef close(self):
        self.flush_all()

    cpdef flush_all(self):
        self.flush()
        self._output.flush()

    cpdef int _refresh_buffer(self) except -1:
        """Control the buffer size of _encoder. Flush if necessary"""
        if self._encoder.position() > self._buffer_size:
            self.flush()

    @property
    def n_bytes(self):
        return self._n_total + self._encoder.position()

    def __len__(self):
        return self.n_bytes

    cdef int _write_tag(self, int field_num, int wire_type) except + nogil:
        return self._encoder.append_tag(field_num, wire_type)

    cdef int _write_raw_long(self, int64_t val) except + nogil:
        return self._encoder.append_sint64(val)

    cdef int _write_raw_int(self, int32_t val) except + nogil:
        return self._encoder.append_sint32(val)

    cdef int _write_raw_uint(self, uint32_t val) except + nogil:
        return self._encoder.append_uint32(val)

    cdef int _write_raw_bool(self, bint val) except + nogil:
        return self._encoder.append_bool(val)

    cdef int _write_raw_float(self, float val) except + nogil:
        return self._encoder.append_float(val)

    cdef int _write_raw_double(self, double val) except + nogil:
        return self._encoder.append_double(val)

    cdef int _write_raw_string(self, const char *ptr, uint32_t size) except + nogil:
        return self._encoder.append_string(ptr, size)


cdef class BaseRecordWriter(ProtobufRecordWriter):

    def __init__(self, object schema, object out, encoding='utf-8'):
        self._encoding = encoding
        self._is_utf8 = encoding == "utf-8"
        self._schema = schema
        self._columns = self._schema.columns
        self._schema_snapshot = self._schema.build_snapshot()

        self._crc_c = Checksum()
        self._crccrc_c = Checksum()
        self._curr_cursor_c = 0
        self._n_columns = len(self._columns)
        self._mills_converter = CMillisecondsConverter()
        self._mills_converter_utc = CMillisecondsConverter(local_tz=False)
        self._to_days = utils.to_days

        super(BaseRecordWriter, self).__init__(out)

    @property
    def _crc(self):
        return self._crc_c

    @property
    def _crccrc(self):
        return self._crccrc_c

    cpdef write(self, BaseRecord record):
        cdef:
            int n_record_fields
            int pb_index
            int i
            int data_type_id
            int checksum
            object val
            object data_type

        n_record_fields = len(record)

        if n_record_fields > self._n_columns:
            raise IOError('record fields count is more than schema.')

        for i in range(min(n_record_fields, self._n_columns)):
            if self._schema_snapshot._col_is_partition[i]:
                continue

            val = record._get(i)
            if val is None:
                continue

            pb_index = i + 1
            self._crc_c.c_update_int(pb_index)

            data_type_id = self._schema_snapshot._col_type_ids[i]
            data_type = None
            if data_type_id >= 0 and data_type_to_wired_type[data_type_id] != -1:
                self._write_tag(pb_index, data_type_to_wired_type[data_type_id])
            else:
                data_type = self._schema_snapshot._col_types[i]
                if isinstance(data_type, (types.Array, types.Map, types.Struct)):
                    self._write_tag(pb_index, WIRETYPE_LENGTH_DELIMITED)
                else:
                    raise IOError('Invalid data type: %s' % data_type)

            self._write_field(val, data_type_id, data_type)

        self._refresh_buffer()

        checksum = <int32_t>self._crc_c.getvalue()
        self._write_tag(WIRE_TUNNEL_END_RECORD, WIRETYPE_VARINT)
        self._write_raw_uint(<uint32_t>checksum)
        self._crc_c.c_reset()
        self._crccrc_c.c_update_int(checksum)
        self._curr_cursor_c += 1

    cdef void _write_bool(self, bint data) except + nogil:
        self._crc_c.c_update_bool(data)
        self._write_raw_bool(data)

    cdef void _write_long(self, int64_t data) except + nogil:
        self._crc_c.c_update_long(data)
        self._write_raw_long(data)

    cdef void _write_float(self, float data) except + nogil:
        self._crc_c.c_update_float(data)
        self._write_raw_float(data)

    cdef void _write_double(self, double data) except + nogil:
        self._crc_c.c_update_double(data)
        self._write_raw_double(data)

    @cython.nonecheck(False)
    cdef _write_string(self, object data):
        cdef bytes bdata
        if type(data) is bytes:
            bdata = data
        elif self._is_utf8 and type(data) is unicode:
            bdata = (<unicode> data).encode("utf-8")
        elif isinstance(data, unicode):
            bdata = data.encode(self._encoding)
        else:
            bdata = bytes(data)
        self._crc_c.c_update(bdata, len(bdata))
        self._write_raw_string(bdata, len(bdata))

    cdef _write_timestamp_base(self, object data, bint ntz):
        cdef:
            object py_datetime = data.to_pydatetime(warn=False)
            long l_val
            int nanosecs
            CMillisecondsConverter converter

        if ntz:
            converter = self._mills_converter_utc
        else:
            converter = self._mills_converter

        l_val = converter.to_milliseconds(py_datetime) // 1000
        nanosecs = data.microsecond * 1000 + data.nanosecond
        self._crc_c.c_update_long(l_val)
        self._write_raw_long(l_val)
        self._crc_c.c_update_int(nanosecs)
        self._write_raw_int(nanosecs)

    cdef _write_timestamp(self, object data):
        return self._write_timestamp_base(data, False)

    cdef _write_timestamp_ntz(self, object data):
        return self._write_timestamp_base(data, True)

    cdef _write_interval_day_time(self, object data):
        cdef:
            long l_val
            int nanosecs

        l_val = data.days * 24 * 3600 + data.seconds
        nanosecs = data.microseconds * 1000 + data.nanoseconds
        self._crc_c.c_update_long(l_val)
        self._write_raw_long(l_val)
        self._crc_c.c_update_int(nanosecs)
        self._write_raw_int(nanosecs)

    cdef _write_field(self, object val, int data_type_id, object data_type):
        cdef int64_t l_val

        if data_type_id == BIGINT_TYPE_ID:
            self._write_long(val)
        elif data_type_id == STRING_TYPE_ID:
            self._write_string(val)
        elif data_type_id == DOUBLE_TYPE_ID:
            self._write_double(val)
        elif data_type_id == FLOAT_TYPE_ID:
            self._write_float(val)
        elif data_type_id == DECIMAL_TYPE_ID:
            self._write_string(str(val))
        elif data_type_id == BINARY_TYPE_ID:
            self._write_string(val)
        elif data_type_id == BOOL_TYPE_ID:
            self._write_bool(val)
        elif data_type_id == DATETIME_TYPE_ID:
            l_val = self._mills_converter.to_milliseconds(val)
            self._write_long(l_val)
        elif data_type_id == DATE_TYPE_ID:
            self._write_long(self._to_days(val))
        elif data_type_id == TIMESTAMP_TYPE_ID:
            self._write_timestamp(val)
        elif data_type_id == TIMESTAMP_NTZ_TYPE_ID:
            self._write_timestamp_ntz(val)
        elif data_type_id == INTERVAL_DAY_TIME_TYPE_ID:
            self._write_interval_day_time(val)
        elif data_type_id == INTERVAL_YEAR_MONTH_TYPE_ID:
            self._write_long(val.total_months())
        elif data_type_id == JSON_TYPE_ID:
            self._write_string(json.dumps(val))
        else:
            if isinstance(data_type, types.Array):
                self._write_raw_uint(len(val))
                self._write_array(val, data_type.value_type)
            elif isinstance(data_type, types.Map):
                self._write_raw_uint(len(val))
                self._write_array(compat.lkeys(val), data_type.key_type)
                self._write_raw_uint(len(val))
                self._write_array(compat.lvalues(val), data_type.value_type)
            elif isinstance(data_type, types.Struct):
                self._write_struct(val, data_type)
            else:
                raise IOError('Invalid data type: %s' % data_type)

    cdef _write_array(self, object data, object data_type):
        cdef int data_type_id = data_type._type_id
        for value in data:
            if value is None:
                self._write_raw_bool(True)
            else:
                self._write_raw_bool(False)
                self._write_field(value, data_type_id, data_type)

    cdef _write_struct(self, object data, object data_type):
        cdef tuple tp_val
        if isinstance(data, dict):
            vals = [None] * len(data)
            for idx, key in enumerate(data_type.field_types.keys()):
                vals[idx] = data[key]
            tp_val = tuple(vals)
        else:
            tp_val = <tuple>data

        for value, field_type in zip(tp_val, data_type.field_types.values()):
            if value is None:
                self._write_raw_bool(True)
            else:
                self._write_raw_bool(False)
                self._write_field(value, field_type._type_id, field_type)

    @property
    def count(self):
        return self._curr_cursor_c

    @property
    def _curr_cursor(self):
        return self._curr_cursor_c

    @_curr_cursor.setter
    def _curr_cursor(self, value):
        self._curr_cursor_c = value

    cpdef _write_finish_tags(self):
        self._write_tag(WIRE_TUNNEL_META_COUNT, WIRETYPE_VARINT)
        self._write_raw_long(self._curr_cursor_c)
        self._write_tag(WIRE_TUNNEL_META_CHECKSUM, WIRETYPE_VARINT)
        self._write_raw_uint(<uint32_t>self._crccrc_c.c_getvalue())

    cpdef close(self):
        self._write_finish_tags()
        super(BaseRecordWriter, self).close()
        self._curr_cursor_c = 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # if an error occurs inside the with block, we do not commit
        if exc_val is not None:
            return
        self.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .arithmetic import UnaryOp
from .expressions import *
from . import utils
from .. import types


class DatetimeOp(UnaryOp):
    __slots__ = ()

    def accept(self, visitor):
        return visitor.visit_datetime_op(self)


class Date(DatetimeOp):
    __slots__ = ()


class Time(DatetimeOp):
    __slots__ = ()


class Year(DatetimeOp):
    __slots__ = ()


class Month(DatetimeOp):
    __slots__ = ()


class Day(DatetimeOp):
    __slots__ = ()


class Hour(DatetimeOp):
    __slots__ = ()


class Minute(DatetimeOp):
    __slots__ = ()


class Second(DatetimeOp):
    __slots__ = ()


class MilliSecond(DatetimeOp):
    __slots__ = ()


class MicroSecond(DatetimeOp):
    __slots__ = ()


class Week(DatetimeOp):
    __slots__ = ()


class WeekOfYear(DatetimeOp):
    __slots__ = ()


class WeekDay(DatetimeOp):
    __slots__ = ()


class UnixTimestamp(DatetimeOp):
    __slots__ = ()


class DayOfYear(DatetimeOp):
    __slots__ = ()


class IsMonthStart(DatetimeOp):
    __slots__ = ()


class IsMonthEnd(DatetimeOp):
    __slots__ = ()


class IsYearStart(DatetimeOp):
    __slots__ = ()


class IsYearEnd(DatetimeOp):
    __slots__ = ()


class DTScalar(Int64Scalar):
    __slots__ = ()


class YearScalar(DTScalar):
    __slots__ = ()


class MonthScalar(DTScalar):
    __slots__ = ()


class DayScalar(DTScalar):
    __slots__ = ()


class HourScalar(DTScalar):
    __slots__ = ()


class MinuteScalar(DTScalar):
    __slots__ = ()


class SecondScalar(DTScalar):
    __slots__ = ()


class MilliSecondScalar(DTScalar):
    __slots__ = ()


class MicroSecondScalar(DTScalar):
    __slots__ = ()


class Strftime(DatetimeOp):
    _args = '_input', '_date_format'
    _add_args_slots = False

    def _init(self, *args, **kwargs):
        self._init_attr('_date_format', None)

        super(Strftime, self)._init(*args, **kwargs)

        if self._date_format is not None and not isinstance(self._date_format, Scalar):
            self._date_format = Scalar(_value=self._date_format)

    @property
    def date_format(self):
        return self._date_format.value


def datetime_op(expr, output_expr_cls, output_type=None, **kwargs):
    output_type = output_type or types.datetime

    if isinstance(expr, (DatetimeSequenceExpr, DatetimeScalar)):
        is_sequence = isinstance(expr, DatetimeSequenceExpr)

        if is_sequence:
            return output_expr_cls(_data_type=output_type, _input=expr, **kwargs)
        else:
            return output_expr_cls(_value_type=output_type, _input=expr, **kwargs)


@property
def _date(expr):
    return datetime_op(expr, Date)


@property
def _time(expr):
    return datetime_op(expr, Time)


@property
def _year(expr):
    return datetime_op(expr, Year, output_type=types.int64)


@property
def _month(expr):
    return datetime_op(expr, Month, output_type=types.int64)


@property
def _day(expr):
    return datetime_op(expr, Day, output_type=types.int64)


@property
def _hour(expr):
    return datetime_op(expr, Hour, output_type=types.int64)


@property
def _minute(expr):
    return datetime_op(expr, Minute, output_type=types.int64)


@property
def _second(expr):
    return datetime_op(expr, Second, output_type=types.int64)


@property
def _millisecond(expr):
    return datetime_op(expr, MilliSecond, output_type=types.int64)


@property
def _microsecond(expr):
    return datetime_op(expr, MicroSecond, output_type=types.int64)


@property
def _week(expr):
    return datetime_op(expr, Week, output_type=types.int64)


@property
def _weekofyear(expr):
    return datetime_op(expr, WeekOfYear, output_type=types.int64)


@property
def _weekday(expr):
    return datetime_op(expr, WeekDay, output_type=types.int64)


@property
def _dayofyear(expr):
    return datetime_op(expr, DayOfYear, output_type=types.int64)


@property
def _unix_timestamp(expr):
    return datetime_op(expr, UnixTimestamp, output_type=types.int64)


@property
def _is_month_start(expr):
    return datetime_op(expr, IsMonthStart, output_type=types.boolean)


@property
def _is_month_end(expr):
    return datetime_op(expr, IsMonthEnd, output_type=types.boolean)


@property
def _is_year_start(expr):
    return datetime_op(expr, IsYearStart, output_type=types.boolean)


@property
def _is_year_end(expr):
    return datetime_op(expr, IsYearEnd, output_type=types.boolean)


def _strftime(expr, date_format):
    """
    Return formatted strings specified by date_format,
    which supports the same string format as the python standard library.
    Details of the string format can be found in python string format doc

    :param expr:
    :param date_format: date format string (e.g. “%Y-%m-%d”)
    :type date_format: str
    :return:
    """

    return datetime_op(expr, Strftime, output_type=types.string,
                       _date_format=date_format)


def year(val):
    return YearScalar(_value=val, _value_type=types.int64)


def month(val):
    return MonthScalar(_value=val, _value_type=types.int64)


def day(val):
    return DayScalar(_value=val, _value_type=types.int64)


def hour(val):
    return HourScalar(_value=val, _value_type=types.int64)


def minute(val):
    return MinuteScalar(_value=val, _value_type=types.int64)


def second(val):
    return SecondScalar(_value=val, _value_type=types.int64)


def millisecond(val):
    return MilliSecondScalar(_value=val, _value_type=types.int64)


def microsecond(val):
    return MicroSecondScalar(_value=val, _value_type=types.int64)


_date_methods = dict(
    year=_year,
    month=_month,
    day=_day,
    week=_week,
    weekofyear=_weekofyear,
    dayofweek=_weekday,
    weekday=_weekday,
    dayofyear=_dayofyear,
    is_month_start=_is_month_start,
    is_month_end=_is_month_end,
    is_year_start=_is_year_start,
    is_year_end=_is_year_end,
)
_datetime_methods = dict(
    date=_date,
    time=_time,
    hour=_hour,
    minute=_minute,
    second=_second,
    microsecond=_microsecond,
    unix_timestamp=_unix_timestamp,
    strftime=_strftime
)
_datetime_methods.update(_date_methods)

utils.add_method(DateSequenceExpr, _date_methods)
utils.add_method(DatetimeSequenceExpr, _datetime_methods)
utils.add_method(DateScalar, _date_methods)
utils.add_method(DatetimeScalar, _datetime_methods)
